# import sksurv.metrics
# import os
# import random
import numpy as np
import pandas as pd
import time
# from time import strftime, gmtime
# import matplotlib.pyplot as plt
# import seaborn as sns
# sns.set()
# from lifelines import KaplanMeierFitter
# from lifelines.utils import concordance_index

from sklearn.model_selection import StratifiedKFold
from sklearn.model_selection import ParameterGrid

from .. import constants as cnt
from .. import metrics as metr

###### 02/10/22

def generate_sample(X, y, folds):
    """
    Generate cross-validate samples with StratifiedKFold.

    Parameters
    ----------
    X : Pandas dataframe
        Contain input features of events.
    y : structured array
        Contain censuring flag and time of events.
    folds : int
        Quantity of cross-validate folds.

    Yields
    ------
    X_train : Pandas dataframe
        Contain input features of train sample.
    y_train : array-like
        Contain censuring flag and time of train sample.
    X_test : Pandas dataframe
        Contain input features of test sample.
    y_test : array-like
        Contain censuring flag and time of test sample.
    bins : array-like
        Points of timeline.

    """
    skf = StratifiedKFold(n_splits=folds)
    for train_index, test_index in skf.split(X, y[cnt.CENS_NAME]):
        X_train, X_test = X.loc[train_index, :], X.loc[test_index, :]
        y_train, y_test = y[train_index], y[test_index]
        bins = cnt.get_bins(time = y_train[cnt.TIME_NAME], cens = y_train[cnt.CENS_NAME])
        y_train[cnt.TIME_NAME] = np.clip(y_train[cnt.TIME_NAME], bins.min()-1, bins.max()+1)
        y_test[cnt.TIME_NAME] = np.clip(y_test[cnt.TIME_NAME], bins.min(), bins.max())
        yield (X_train, y_train, X_test, y_test, bins)
    pass
    

def crossval_param(method, X, y, folds, metrics_names = ['CI']):
    """
    Return function, which on sample X, y apply cross-validation and calculate 
    metrics on each folds. 

    Parameters
    ----------
    method : object
        Must have methods for fitting, predicting time, hazard and survival func
            
    X : Pandas dataframe
        Contain input features of events.
    y : structured array
        Contain censuring flag and time of events.
    folds : int
        Quantity of cross-validate folds.
    metrics_names : TYPE, optional
        DESCRIPTION. The default is ['CI'].

    Returns
    -------
    functions
        Recieve hyperparameters and return list of metrics arrays.
        Allow to use in ParameterGrid.

    """
    def f(**kwargs):
        metr_lst = []
        for X_train, y_train, X_test, y_test, bins in generate_sample(X, y, folds):
            if method.__name__.find('CRAID') != -1: #TODO replace to isinstance 
                est = method(**kwargs)
                est.fit(X_train, y_train)
                pred_surv = est.predict_at_times(X_test, bins = bins, mode = "surv")
                pred_time = est.predict(X_test, target = cnt.TIME_NAME)
                pred_haz = est.predict_at_times(X_test, bins = bins, mode = "hazard")
            else:
                X_train = X_train.fillna(0).replace(np.nan, 0)
                X_test = X_test.fillna(0).replace(np.nan, 0)
                
                est = method(**kwargs).fit(X_train, y_train)
                survs = est.predict_survival_function(X_test)
                hazards = est.predict_cumulative_hazard_function(X_test)
                pred_surv = np.array(list(map(lambda x: x(bins), survs)))
                pred_haz = np.array(list(map(lambda x: x(bins), hazards)))
                pred_time = -1*est.predict(X_test)
            
            metr_lst.append(np.array([metr.METRIC_DICT[metr_name](y_train, y_test, 
                                                        pred_time, pred_surv, pred_haz, bins) 
                                  for metr_name in metrics_names]))
        return np.vstack(metr_lst)
    return f


class Experiments(object):
    """
    Class recieves methods, metrics and grids, 
          produces cross-validation experiments,
          stores table of results : name, params, time, metrics (by sample and mean)

    Attributes
    ----------
    methods : list
        Must have predicting methods according to metrics:
            IBS - survival func
            IAUC - cumulative hazard func
            CI - occured time
            CI_CENS - occured time
    methods_grid : list
        Each grid is dictionary: key - param name, values - list
    metrics : list
        Each metric is string, which must be in METRIC_DICT
    is_table : boolean
        Flag of calculation ending
    folds : int
        Quantity of cross-validate folds.
    except_stop : str
        Mode of ending because of exception
        "all" - continue experiments
        else - stop experiments with current method
    dataset_name : str
        Unique name of current dataset (used for saving)

    Methods
    -------
    
    add_method : append method and its grid
    set_metrics : check and set list of metric name 
    run : start experiments with data X, y
    get_best_results : choose for each method best params by metric and aggreg 
    save : export table as xlsx
    
    """
    def __init__(self, folds = 5, except_stop = "all", dataset_name = "NONE_NAME"):
        self.methods = []
        self.methods_grid = []
        self.metrics = ["CI"]
        self.is_table = False
        self.folds = folds
        self.except_stop = except_stop
        self.dataset_name = dataset_name
        
    def add_method(self, method, grid):
        self.methods.append(method)
        self.methods_grid.append(grid)
        
    def set_metrics(self, lst_metric):
        self.metrics = []
        for metr_name in lst_metric:
            if metr_name in metr.METRIC_DICT:
                self.metrics.append(metr_name)
            else:
                print("METRIC %s IS NOT DEFINED" % (metr_name))
    
    def run(self, X, y, dir_path = None, verbose = 0):
        self.result_table = pd.DataFrame([], columns = ["METHOD", "PARAMS", "TIME"] + self.metrics)
        
        for method, grid in zip(self.methods, self.methods_grid):
            cv_method = crossval_param(method, X, y, self.folds, self.metrics)
            try:
                for p in ParameterGrid(grid):
                    start_time = time.time()
                    cv_metr = cv_method(**p)
                    full_time = time.time() - start_time
                    curr_dict = {"METHOD": method.__name__, "CRIT": p.get("criterion",""),
                                 "PARAMS": str(p), "TIME": full_time}
                    cv_metr = {m: cv_metr[:,i] for i, m in enumerate(self.metrics)}
                    curr_dict.update(cv_metr) #dict(zip(self.metrics, cv_metr))
                    self.result_table = self.result_table.append(curr_dict, ignore_index = True)
                    if verbose > 0:
                        print("EXECUTION TIME OF %s: %s" % (method.__name__, full_time), 
                              {k:np.mean(v) for k,v in cv_metr.items()})
            except KeyboardInterrupt:
                print("HANDELED KeyboardInterrupt")
                break
            except Exception as e:
                print("Method: %s, Param: %s finished with except '%s'" % (method.__name__, str(p), e))
                if self.except_stop == "all":
                    break
        for m in self.metrics:
            self.result_table["%s_mean" % (m)] = self.result_table[m].apply(np.mean)
        self.is_table = True
        if not(dir_path is None):
            # add_time = strftime("%H:%M:%S", gmtime(time.time()))
            self.save(dir_path)
    
    def get_result(self):
        return self.result_table
    
    def get_best_results(self, by_metric, choose = "max"):
        if not(by_metric in self.metrics):
            return None
        
        best_table = pd.DataFrame([], columns = self.result_table.columns)
        for method in self.result_table['METHOD'].unique():
            sub_table = self.result_table[self.result_table["METHOD"] == method]
            if sub_table.shape[0] == 0:
                continue
            if choose == "max":
                best_row = sub_table.loc[sub_table[by_metric].apply(np.mean).idxmax()]
            else:
                best_row = sub_table.loc[sub_table[by_metric].apply(np.mean).idxmin()]
            best_table = best_table.append(dict(best_row), ignore_index = True)
        return best_table
    
    def save(self, dir_path):
        # save_table = self.result_table.copy()
        # if dir_path.find(".csv") != -1:
        #     save_table.to_csv(dir_path + dataset)
        # else: #to_excel
        self.result_table.to_excel("%s_FULL_TABLE.xlsx" % (dir_path + self.dataset_name), index = False)
            
    # def plot_results(self, dir_path):
    #     df = self.result_table.copy()
    #     df['METHOD'] = df.apply(lambda x: x["METHOD"].replace("CRAID", "Tree(%s)" %(x['CRIT'])),
    #                                  axis = 1)
    #     for m in self.metrics:
    #         fig, axs = plt.subplots(1, figsize=(6, 9))
    #         plt.title("%s %s" % (self.dataset_name, m))
    #         plt.boxplot(df[m], labels = df['METHOD'])
    #         plt.xticks(rotation=90)
    #         plt.savefig(dir_path + self.dataset_name + "%s_boxplot.png" %(m))
    #         plt.close(fig)


import numpy as np
import pandas as pd
import re
import pickle 
import random
from os.path import dirname, join
from ..constants import TIME_NAME, CENS_NAME, get_y
random.seed(10)

sign_gbsg = ['htreat','age','menostat','tumsize','tumgrad','posnodal','prm','esm']
categ_gbsg = ['htreat', 'menostat', 'tumgrad']

sign_pbc = ['trt','age', 'sex', 'ascites','hepato', 'spiders', 'edema', 'bili', 'chol', 
            'albumin', 'copper', 'alk', 'ast', 'trig', 'platelet', 'protime', 'stage']
categ_pbc = ['trt', 'sex', 'ascites','hepato', 'spiders']

def save_pickle(obj, path):
    file_pi = open(path, 'wb') 
    pickle.dump(obj, file_pi, pickle.HIGHEST_PROTOCOL)
    
def load_pickle(path):
    return pickle.load(open(path, 'rb'))

def load_gbsg_dataset():
    dir_env = join(dirname(__file__), "data")
    df = pd.read_csv(join(dir_env, 'GBSG.csv'))
    df = df.rename({"rfst":TIME_NAME,"cens":CENS_NAME},axis = 1)
    
    y = get_y(df[CENS_NAME], df[TIME_NAME])
    X = df.loc[:, sign_gbsg]
    return X, y, sign_gbsg, categ_gbsg, []

def load_pbc_dataset():
    dir_env = join(dirname(__file__), "data")
    df = pd.read_csv(join(dir_env, 'pbc.csv'))
    df = df.rename({"time":TIME_NAME,"status":CENS_NAME, "alk.phos":"alk"},axis = 1)
    df[CENS_NAME] = np.array(df[CENS_NAME] > 1, dtype = int)
    df['sex'] = df['sex'].map({'f':1, 'm':0})
    
    y = get_y(df[CENS_NAME], df[TIME_NAME])
    X = df.loc[:, sign_pbc]
    return X, y, sign_pbc, categ_pbc, []

def load_wuhan_dataset(invert_death = False):
    dir_env = join(dirname(__file__), "data")
    df = pd.read_excel(join(dir_env, 'covid_train.xlsx'))
    df['PATIENT_ID'] = df['PATIENT_ID'].fillna(method = 'ffill')
    columns_no_agg = ['RE_DATE','age','gender','Admission time','Discharge time','outcome']
    #columns_agg = list(set(df.columns) - set(columns_no_agg))
    df_agg = df.groupby('PATIENT_ID').agg(list)
    for c in df_agg.columns:
        if c in columns_no_agg:
            df_agg[c] = df_agg[c].apply(lambda x: x[0])
        else:
            df_agg['mean_' + c] = df_agg[c].apply(np.nanmean)
            df_agg['min_' + c] = df_agg[c].apply(np.nanmin)
            df_agg['max_' + c] = df_agg[c].apply(np.nanmax)
            df_agg = df_agg.drop(c, axis = 1)
    df_agg['time'] = df_agg.loc[:,['Admission time','Discharge time']].apply(lambda x: (x['Discharge time'] - x['Admission time']).days, axis = 1)
    df_f = df_agg[df_agg['time'] == df_agg['time']]
    df_f = df_f.drop(['RE_DATE', 'Admission time', 'Discharge time'], axis = 1)
    df_f = df_f.rename({'outcome':CENS_NAME, 'time':TIME_NAME}, axis = 1)
    df_f = df_f.rename({c:re.sub('[^A-Za-z0-9_]', '_', c) for c in df_f.columns}, axis = 1)
    df_f = df_f.reset_index(drop=True)
    
    categ_covid = []
    sign_covid = list(set(df_f) - {CENS_NAME, TIME_NAME})
    
    if invert_death:
        df_f[CENS_NAME] = 1 - df_f[CENS_NAME]
    y = get_y(df_f[CENS_NAME], df_f[TIME_NAME])
    X = df_f.loc[:, sign_covid]
    
    return X, y, sign_covid, categ_covid, []


import numpy as np
import pandas as pd
import re
from os.path import dirname, join
from dateutil import parser
from ..constants import TIME_NAME, CENS_NAME, get_y

schemes_list = ['light_1', 'light_2',
        'light_3', 'middle_1', 'middle_2', 'middle_3', 'middle_4', 'middle_5',
        'middle_6', 'middle_7', 'middle_8', 'hard_1', 'hard_2', 'hard_3',
        'hard_4', 'hard_5', 'cyto_1', 'cyto_2', 'cyto_3', 'cyto_4', 'cyto_5',
        'cyto_6',
        'MOS_AM', 'MOS_AM_AZ', 'N_MOS_AM_AZ', 'MOS_AM_LVF', 'N_MOS_AM_LVF',
        'MOS_CFTR', 'MOS_CFT', 'MOS_CFP', 'MOS_CFP_LVF', 'N_MOS_CFP_LVF',
        'MOS_MOK', 'MOS_PIP', 'MOS_LVF_MOK', 'mos_light_1', 'mos_light_2',
        'mos_middle_1', 'mos_middle_2', 'mos_middle_3', 'mos_middle_4',
        'mos_hard_1', 'mos_hard_2', 'mos_hard_3', 'mos_cyto_1', 'mos_cyto_2',
#        'cheme_therap_nmg_hor_inter', 'cheme_therap_nmg_hor',
#        'cheme_therap_nmg_inter', 'dexamethasone', 'methylprednisolone',
#        'dex_meth'
        ]

important = [TIME_NAME, CENS_NAME, 'SUM_M']

categ_cyto = ['has_odyshka_ili_zatrudnennoe_dyhanie','has_oschussh_zalozhen_v_grudnoi_kletke',
              'has_slabost_ili_lomota','Пол','IGG_DEF','IGM_DEF','PCR_N',
              'сахарный диабет', 'ожирение','ишемическая болезнь сердца',
              'has_kashel','kashel_type','NIT','resultat_KT'] + schemes_list
def_categ = {
      'osmotr_tyazhest':{'без симптомов': 0 ,'легкая': 3,'средняя':5, 'тяжелая':7},
      'has_odyshka_ili_zatrudnennoe_dyhanie':{'отсутствует': 0 ,'присутствует': 1},
      'has_oschussh_zalozhen_v_grudnoi_kletke':{'отсутствует': 0 ,'присутствует': 1},
      'has_slabost_ili_lomota':{'отсутствует': 0 ,'присутствует': 1},
      'Пол':{'Мужской': 0 ,'Женский': 1, 'м': 0, 'ж': 1,'Нет данных': np.nan, 0: 0, 1: 1},
      'PCR_N':{'НЕ ОБНАРУЖЕНО':0,'ОБНАРУЖЕНО':1},
      'has_kashel':{'отсутствует': 0 ,'присутствует': 1},
      'kashel_type':{'сухой': 0, 'с мокротой': 1},
      'Группа риска':{'Хроники': 0 ,'Беременные': 1, 'Старше 65': 2}
}

def_categ_full = {
      'osmotr_tyazhest':{'без симптомов': 0 ,'легкая': 3,'средняя':5, 'тяжелая':7},
      'has_odyshka_ili_zatrudnennoe_dyhanie':{'отсутствует': 0 ,'присутствует': 1, 0: 0, 1: 1},
      'has_oschussh_zalozhen_v_grudnoi_kletke':{'отсутствует': 0 ,'присутствует': 1, 0: 0, 1: 1},
      'has_slabost_ili_lomota':{'отсутствует': 0 ,'присутствует': 1, 0: 0, 1: 1},
      'Пол':{'Мужской': 0 ,'Женский': 1, 'м': 0, 'ж': 1,'Нет данных': np.nan, 0: 0, 1: 1},
      'PCR_N':{'НЕ ОБНАРУЖЕНО':0,'ОБНАРУЖЕНО':1, 0: 0, 1: 1},
      'has_kashel':{'отсутствует': 0 ,'присутствует': 1, 0: 0, 1: 1},
      'kashel_type':{'сухой': 0, 'с мокротой': 1, 0: 0, 1: 1},
      'Группа риска':{'Хроники': 0 ,'Беременные': 1, 'Старше 65': 2, 0: 0, 1: 1, 2:2}
}

sign = [#'osmotr_tyazhest',
        'days_f','days_s', 'spo2','chdd','has_odyshka_ili_zatrudnennoe_dyhanie','temperatura_tela_value',
        'has_oschussh_zalozhen_v_grudnoi_kletke', 'has_slabost_ili_lomota','Пол','возраст',
        'IGG_N','IGM_N','IGG_DEF','IGM_DEF','PCR_N',
        'сахарный диабет', 'ожирение','ишемическая болезнь сердца',
        'Определение белков острой фазы С-реактивный белок',
        'has_kashel','kashel_type',
        'D-димер', 'Исследование ферритина',
        'NIT', 'АСТ', 'АЛТ', 'URO', 'Определение креатинина', 
        'GRA#','Определение калия общего',
        'по Вестергрену','по Панченкову',
        'MON#',
        'HGB','MPV','HCT','PLT', 'PDW',
        'LYM#',
        'GLU','WBC', 'RBC',
        'RDW',
        'Абсолютное количество эозинофилов',
        'Абсолютное количество базофилов',
        'Абсолютное количество нейтрофилов',
        'resultat_KT'
        ] + schemes_list

therap_antic = "Гепарин Далтепарин Надропарин Эноксапарин Парнапарин Бемипарин Фондапаринукс Ривароксабан Апиксабан"
prevent_antic = therap_antic
glucoco = "Преднизолон Дексаметазон Метилпреднизолон Гидрокортизон"

mapp_sch = {
    "light_1":"Гидроксихлорохин",
    "light_2":"Мефлохин",
    "light_3":"ИФН Умифеновир",
    "middle_1":"Фавипиравир Барицитиниб Тофацитиниб",
    "middle_2":"Гидроксихлорохин Азитромицин Барицитиниб Тофацитиниб",
    "middle_3":"Мефлохин Азитромицин Барицитиниб Тофацитиниб",
    "middle_4":"Лопинавир Ритонавир ИФН Барицитиниб Тофацитиниб",
    "middle_5":"Фавипиравир Олокизумаб",
    "middle_6":"Гидроксихлорохин Азитромицин Олокизумаб",
    "middle_7":"Мефлохин Азитромицин Олокизумаб",
    "middle_8":"Лопинавир Ритонавир ИФН Олокизумаб",
    "hard_1":"Фавипиравир Тоцилизумаб Сарилумаб",
    "hard_2":"Гидроксихлорохин Азитромицин Тоцилизумаб Сарилумаб",
    "hard_3":"Мефлохин Азитромицин Тоцилизумаб Сарилумаб",
    "hard_4":"Лопинавир Ритонавир ИФН Тоцилизумаб Сарилумаб",
    "hard_5":"Лопинавир Ритонавир Гидроксихлорохин Тоцилизумаб Сарилумаб",
    "cyto_1":"Метилпреднизолон Тоцилизумаб Сарилумаб",
    "cyto_2":"Дексаметазон Тоцилизумаб Сарилумаб",
    "cyto_3":"Метилпреднизолон Канакинумаб",
    "cyto_4":"Дексаметазон Канакинумаб",
    "cyto_5":"Метилпреднизолон Дексаметазон",
    "cyto_6":"Тоцилизумаб Сарилумаб Канакинумаб",
    "MOS_AM":"Амоксициллин",
    "MOS_AM_AZ":"Ампициллин Азитромицин",   #Антибиотик: Ампициллин
    "N_MOS_AM_AZ":"Азитромицин",
    "MOS_AM_LVF":"Ампициллин Левофлоксацин",#Антибиотик: Ампициллин
    "N_MOS_AM_LVF":"Левофлоксацин",
    "MOS_CFTR":"Цефтриаксон",               #Антибиотик Цефтриаксон
    "MOS_CFT":"Цефотаксим",                 #Антибиотик Цефотаксим
    "MOS_CFP":"Цефепим",                    #Антибиотик Цефепим
    "MOS_CFP_LVF":"Цефепим Левофлоксацин",  #Антибиотик Цефепим
    "N_MOS_CFP_LVF":"Левофлоксацин",
    "MOS_MOK":"Моксифлоксацин",
    "MOS_PIP":"Пиперациллин",               #Антибиотик Пиперациллин
    "MOS_LVF_MOK":"Левофлоксацин Моксифлоксацин",
'mos_light_1':"Фавипиравир Риамиловир Гидроксихлорохин ",
'mos_light_2':"Фавипиравир Риамиловир Гидроксихлорохин " + therap_antic,
'mos_middle_1':"Фавипиравир Риамиловир Гидроксихлорохин " + prevent_antic, 
'mos_middle_2':"Фавипиравир Риамиловир Гидроксихлорохин " + prevent_antic,
'mos_middle_3':"Фавипиравир Риамиловир Гидроксихлорохин " + prevent_antic, 
'mos_middle_4':"Фавипиравир Риамиловир Гидроксихлорохин " + prevent_antic,
'mos_hard_1':"Фавипиравир " + prevent_antic + glucoco, 
'mos_hard_2':"Риамиловир " + prevent_antic + glucoco, 
'mos_hard_3':"Гидроксихлорохин "+prevent_antic + glucoco, 
'mos_cyto_1':"Фондапаринукс " + glucoco,
'mos_cyto_2':"Фондапаринукс " + prevent_antic + glucoco
}
map_bool = {'0':'Нет','1':'Да'}

def create_none_ft(fill_df, sign_f, fill_none = False):
    for i in sign_f:
      #fill_df['none_'+i] = fill_df[i].apply(lambda x: int(x != x))
        if i in set(def_categ.keys()):
            # categorical features
            fill_df[i] = fill_df[i].map(def_categ[i])
            print('Categ-contin var:', i)
            if fill_none:
                fill_df[i] = fill_df[i].fillna(fill_df[i].median())
                fill_df[i] = fill_df[i].astype('int')
        else:
            # continuous features
            right_thres = 4
            if fill_none:
                right_thres = 3
                fill_df[i] = fill_df[i].fillna(fill_df[i].median())
                fill_df[i] = fill_df[i].astype('float64')
            len_uniq = len(fill_df[i].unique())
            if len_uniq < 10:
                if len_uniq <= 6 and len_uniq >= right_thres:
                    print('Categorical var:', i, len_uniq)
                    dict_encoder = {v_u:i_u for i_u, v_u in enumerate(np.sort(fill_df[i].unique())) if v_u == v_u}
                    fill_df[i] = fill_df[i].map(dict_encoder)
    return fill_df

def dest_date(x,y, with_neg = False):
    days = np.nan
    try:
        p_1 = parser.parse(x)
        p_2 = parser.parse(y)
        days = (p_1 - p_2).days
        if not with_neg:
            days = abs(days)
    except:
        pass
    return days

def load_covid_dataset():
    dir_env = join(dirname(__file__), "data", "COVID")
    cyto = pd.read_csv(join(dir_env, "cyto_with_schemes.csv"))
    cyto = create_none_ft(cyto, sign)
    cyto[TIME_NAME] = cyto.apply(lambda row: dest_date(row['Дата создания ЭС'], row['Дата закрытия ЭС']),axis = 1)
    cyto[CENS_NAME] = cyto['Причина закрытия ЭС'].apply(lambda x: int(x[0] == 'У') if x == x else 0)
    cyto['SUM_M'] = 0
    sec = cyto.loc[:,sign+important]
    
    repl = {sc:re.sub("[^0-9A-Za-zА-Яа-я_]","",sc) for sc in sec.columns}
    sec = sec.rename(columns=repl)
    sec['idx'] = sec.index 
    sec = sec[sec[TIME_NAME] == sec[TIME_NAME]]
    for sch in schemes_list:
        sec[sch] = sec[sch].fillna(0)
    new_sign = [repl[s] for s in sign]
    new_categ = [repl[s] for s in categ_cyto]
    new_important = [repl[s] for s in schemes_list]
    
    sec = sec.reset_index()
    y = get_y(sec[CENS_NAME], sec[TIME_NAME])
    X = sec.loc[:, new_sign]
    return X, y, new_sign, new_categ, new_important

#### ADD SCHEMES

# def add_schemes():
#     cyto = pd.read_csv(dir_env + 'cyto_train_16.csv')
#     schemes_df = pd.read_csv(dir_env + 'second_all_schemes.csv')
#     simi = pd.read_excel(dir_env + 'simi_uuid.xlsx')
#     cyto_global = cyto.merge(simi, how = 'left', on = 'uuid')
#     cyto_global['kis_uuid'] = cyto_global.apply(lambda x: x['kis_uuid'] if x['kis_uuid'] == x['kis_uuid'] else x['uuid'],axis = 1)
#     schemes_df = schemes_df.rename({'id':'kis_uuid'},axis = 1)
#     schemes_df = schemes_df.loc[:,['kis_uuid']+schemes_list]#.sum(axis = 1).value_counts()
#     cyto_global = cyto_global.merge(schemes_df, how = 'left', on = 'kis_uuid')
#     cyto_global.to_csv(dir_env + 'cyto_with_schemes.csv')


import pandas as pd
import numpy as np
from lifelines.utils import concordance_index
from sklearn.metrics import roc_auc_score

from .. import metrics as metr
from ..tree import CRAID
from .. import constants as cnt

class BaseEnsemble(object):
    """
    Base ensemble of survival decision tree.

    Attributes
    ----------
    size_sample : float
        Size of generated subsample
    n_estimators : int
        Number of base models 
    ens_metric_name : str
        Metric defines quantitve of ensemble
    descend_metr : boolean 
        Flag of descend for ens_metric_name
    bootstrap : boolean
        Flag for using bootstrap sampling (with return)
    tolerance : boolean
        Flag for fitting full ensemble and choosing best submodels
    aggreg : boolean
        Flag of aggregating base responses
    aggreg_func : str
        Function of aggregating (if aggreg)
    tree_kwargs : dict
        Parameters for building base models
    
    models : list 
        Base models of ensemble (for example, CRAID)
    oob : list
        Out of bag sample for each model
    ens_metr : array-like
        Values of ens_metric_name for ensemble with i models
    list_pred_oob : list
        Predictions of ensemble models[:i+1]

    Methods
    -------
    
    update_params : attributes preparation
    fit : build ensemble with X, y data (abstract)
    add_model : updating ensemble with new model and oob
    select_model : remaining fixed models
    tolerance_find_best : iterative method of selecting best subensemble
    
    predict : return values of features, rules or schemes (look at CRAID)
    predict_at_times : return survival or hazard function

    score_oob : calculate metric "ens_metric_name" for ensemble
    
    """
    def __init__(self, size_sample = 0.7,
                 n_estimators = 10,
                 aggreg = True,
                 ens_metric_name = "roc",
                 bootstrap = True,
                 tolerance = True,
                 aggreg_func = "mean",
                 **tree_kwargs):
        self.size_sample = size_sample
        self.n_estimators = n_estimators
        self.aggreg = aggreg
        self.ens_metric_name = ens_metric_name
        self.descend_metr = self.ens_metric_name in metr.DESCEND_METRICS
        self.bootstrap = bootstrap
        self.tolerance = tolerance
        self.tree_kwargs = tree_kwargs
        self.aggreg_func = aggreg_func
    
    def update_params(self):
        self.models = []
        self.oob = []
        self.ens_metr = np.zeros(self.n_estimators)
        self.list_pred_oob = []
        
        if self.size_sample < 1.0:
            self.size_sample = int(self.size_sample*self.X_train.shape[0])
        self.bins = cnt.get_bins(time = self.y_train[cnt.TIME_NAME], 
                                 cens = self.y_train[cnt.CENS_NAME])
        cnt.set_seed(10)
    
    def fit(self):
        pass
    
    def add_model(self, model, x_oob):
        self.models.append(model)
        self.oob.append(x_oob)
        if self.ens_metric_name == "conc":
            self.list_pred_oob.append(model.predict(x_oob, target = cnt.TIME_NAME).to_frame())
        elif self.ens_metric_name == "roc":
            self.list_pred_oob.append(model.predict(x_oob, target = cnt.CENS_NAME).to_frame())
        else:
            pred_surv = pd.DataFrame(model.predict_at_times(x_oob, bins = self.bins, mode = "surv"), index = x_oob.index)
            self.list_pred_oob.append(pred_surv.apply(lambda r: np.array(r), axis=1))
    
    def select_model(self, start, end):
        self.models = self.models[start:end]
        self.oob = self.oob[start:end]
        self.list_pred_oob = self.list_pred_oob[start:end]
        
    def tolerance_find_best(self):
        if self.descend_metr:
            max_index = np.argmin(self.ens_metr)
        else:
            max_index = np.argmax(self.ens_metr)
        self.select_model(0,max_index+1)
        print(self.ens_metr)
    
    def get_aggreg(self, x):
        if self.aggreg_func == 'median':
            return np.median(x, axis = 0)
        elif self.aggreg_func == "mean":
            return np.mean(x, axis = 0)
        return None
    
    def predict_at_times(self, x_test, bins, aggreg = True, mode = "surv"):
        res = []
        for i in range(len(self.models)):
            res.append(self.models[i].predict_at_times(x_test, bins = bins, 
                                                       mode = mode)[np.newaxis, :])
        res = np.vstack(res)
        if aggreg:
            res = self.get_aggreg(res)
        return res
    
    def predict(self, x_test, aggreg = True, **kwargs):
        res = []
        for i in range(len(self.models)):
            res.append(self.models[i].predict(x_test, **kwargs))
        res = np.array(res)
        if aggreg:
            res = self.get_aggreg(res)
        return res
    
    """ SCORES """
    
    def score_oob(self):
        if self.aggreg:
            score = self.aggregate_score_selfoob()
        else:
            score = self.separate_score_oob()
        return np.round(score, 4)
    
    def separate_score_oob(self):
        scores = np.array([])
        for i in range(len(self.models)):
            X_v = self.oob[i]
            if self.ens_metric_name == "conc":
                pred = self.models[i].predict(X_v, target = cnt.TIME_NAME)
                score = concordance_index(X_v[cnt.TIME_NAME], pred)
            else:
                pred = self.models[i].predict_at_times(X_v, bins = self.bins, mode = "surv")
                y_true = cnt.get_y(X_v[cnt.CENS_NAME], X_v[cnt.TIME_NAME])
                score = metr.ibs(self.y_train, y_true, pred, self.bins)
            scores = np.append(scores, score)
        return np.round(np.mean(scores), 4)
        
    def aggregate_score_selfoob(self, bins = None):
        if self.ens_metric_name in ["conc", "ibs"]:
            list_target_time = [oob_[cnt.TIME_NAME].to_frame() for oob_ in self.oob]
            target_time = pd.concat(list_target_time, axis=1).mean(axis = 1)
            
        if self.ens_metric_name in ["roc", "ibs"]:
            list_target_cens = [oob_[cnt.CENS_NAME].to_frame() for oob_ in self.oob]
            target_cens = pd.concat(list_target_cens, axis=1).mean(axis = 1)
            
        if self.ens_metric_name in ["conc", "roc"]:
            pred = pd.concat(self.list_pred_oob, axis=1).mean(axis = 1)
            if self.ens_metric_name == "conc":
                return concordance_index(target_time, pred)
            return roc_auc_score(target_cens, pred)
        
        if self.ens_metric_name == "ibs":
            pred = pd.concat(self.list_pred_oob, axis=1).apply(lambda r: r.mean(axis = 0),axis = 1)
            pred = np.array(pred.to_list())
            y_true = cnt.get_y(target_cens, target_time)
            return metr.ibs(self.y_train, y_true, pred, self.bins)
        return None


import sksurv.metrics
import numpy as np
from numba import njit, jit
from lifelines import KaplanMeierFitter, NelsonAalenFitter
from lifelines.utils import concordance_index

from .constants import TIME_NAME, CENS_NAME

METRIC_DICT = {
    "CI": lambda y_tr, y_tst, pr_time, pr_surv, pr_haz, bins:
        concordance_index(y_tst[TIME_NAME], pr_time),
    "CI_CENS": lambda y_tr, y_tst, pr_time, pr_surv, pr_haz, bins:
        concordance_index(y_tst[TIME_NAME], pr_time, y_tst[CENS_NAME]),
    "IBS": lambda y_tr, y_tst, pr_time, pr_surv, pr_haz, bins:
        ibs(y_tr, y_tst, pr_surv, bins),
    "IAUC": lambda y_tr, y_tst, pr_time, pr_surv, pr_haz, bins:
        iauc(y_tr, y_tst, pr_haz, bins)
}
""" dict: Available metrics in library and its realization """

DESCEND_METRICS = ['ibs', 'IBS']
""" list: Metrics with decreasing quality improvement """

@jit
def get_before(estim, wei):
    return np.square(estim) * wei

@jit
def get_after(estim, prob_cens):
    return np.square(1 - estim) / prob_cens

def ibs(survival_train, survival_test, estimate, times, axis = -1):
    """
    Modified integrated brier score from scikit-survival (add axis)
    Modification: with axis = 0 count ibs for each observation

    Parameters
    ----------
    survival_train : structured array, shape = (n_train_samples,)
        Survival times for training data to estimate the censoring
        distribution from.
        A structured array containing the binary event indicator
        as first field, and time of event or time of censoring as
        second field.
    survival_test : structured array, shape = (n_samples,)
        Survival times of test data.
        A structured array containing the binary event indicator
        as first field, and time of event or time of censoring as
        second field.
    estimate : array-like, shape = (n_samples, n_times)
        Estimated risk of experiencing an event for test data at `times`.
        The i-th column must contain the estimated probability of
        remaining event-free up to the i-th time point.
    times : array-like, shape = (n_times,)
        The time points for which to estimate the Brier score.
    axis : int, optional
        With axis = 0 count ibs for each observation.
        With axis = -1 count mean ibs.
        The default is -1.

    Returns
    -------
    ibs_value : float or array-like
        if axis = 0 return array of ibs
           axis = -1 return float

    """
    test_event, test_time = sksurv.metrics.check_y_survival(survival_test, allow_all_censored = True)
    #estimate, times = _check_estimate_2d(estimate, test_time, times)
    estimate = np.array(estimate)
    if estimate.ndim == 1 and times.shape[0] == 1:
        estimate = estimate.reshape(-1, 1)
    estimate[estimate == -np.inf] = 0
    estimate[estimate == np.inf] = 0
    # fit IPCW estimator
    cens = sksurv.metrics.CensoringDistributionEstimator().fit(survival_train)
    # calculate inverse probability of censoring weight at current time point t.
    prob_cens_t = cens.predict_proba(times)
    prob_cens_t[prob_cens_t == 0] = np.inf
    # calculate inverse probability of censoring weights at observed time point
    prob_cens_y = cens.predict_proba(test_time)
    prob_cens_y[prob_cens_y == 0] = np.inf
    wei = test_event / prob_cens_y
    
    estim_before = get_before(estimate, wei[np.newaxis,:].T)
    estim_after = get_after(estimate, prob_cens_t)
    brier_scores = np.array([np.where(test_time <= t, 
                                      estim_before[:, i], 
                                      estim_after[:, i])
                             for i, t in enumerate(times)])
    if axis == -1:
        brier_scores = np.mean(brier_scores, axis = 1)
        ibs_value = np.trapz(brier_scores, times) / (times[-1] - times[0])
    else:
        ibs_value = np.trapz(brier_scores, times, axis = 0) / (times[-1] - times[0])
    # else: # Return bs in time (for graphics)
    #     brier_scores = np.mean(brier_scores, axis = 1)
        
    return ibs_value

""" EXPORT IAUC """
def iauc(survival_train, survival_test, estimate, times, tied_tol=1e-8):
    """
    Modified integrated AUC (cumulative_dynamic_auc) 
        from scikit-survival (reduce complexity)

    Parameters
    ----------
    survival_train : structured array, shape = (n_train_samples,)
        Survival times for training data to estimate the censoring
        distribution from.
        A structured array containing the binary event indicator
        as first field, and time of event or time of censoring as
        second field.
    survival_test : structured array, shape = (n_samples,)
        Survival times of test data.
        A structured array containing the binary event indicator
        as first field, and time of event or time of censoring as
        second field.
    estimate : array-like, shape = (n_samples, n_times)
        Estimated risk of experiencing an event for test data at `times`.
        The i-th column must contain the estimated probability of
        remaining event-free up to the i-th time point.
    times : array-like, shape = (n_times,)
        The time points for which to estimate the Brier score.
    tied_tol : float, optional, default: 1e-8
        The tolerance value for considering ties.
        If the absolute difference between risk scores is smaller
        or equal than `tied_tol`, risk scores are considered tied.
        
    Returns
    -------
    mean_auc : float
        Summary measure referring to the mean cumulative/dynamic AUC
        over the specified time range `(times[0], times[-1])`.

    """
    if survival_train[CENS_NAME].sum() == 0:
        survival_train[CENS_NAME] = 1
        survival_test[CENS_NAME] = 1 - survival_test[CENS_NAME]
    if survival_test[CENS_NAME].sum() == 0:
        survival_test[CENS_NAME] = 1
    test_event, test_time = sksurv.metrics.check_y_survival(survival_test)
    #estimate, times = _check_estimate_2d(estimate, test_time, times)
    estimate = np.array(estimate)
    n_samples = estimate.shape[0]
    n_times = times.shape[0]
    if estimate.ndim == 1:
        estimate = np.broadcast_to(estimate[:, np.newaxis], (n_samples, n_times))

    # fit and transform IPCW
    cens = sksurv.metrics.CensoringDistributionEstimator()
    cens.fit(survival_train)
    Ghat = cens.predict_proba(test_time[test_event])
    ipcw = np.zeros(test_time.shape[0])
    Ghat[Ghat == 0] = np.inf
    if not((Ghat == 0.0).any()):
        ipcw[test_event] = 1.0 / Ghat
    else:
        ipcw = np.ones(test_time.shape[0])

    # expand arrays to (n_samples, n_times) shape
    test_time = np.broadcast_to(test_time[:, np.newaxis], (n_samples, n_times))
    test_event = np.broadcast_to(test_event[:, np.newaxis], (n_samples, n_times))
    times_2d = np.broadcast_to(times, (n_samples, n_times))
    ipcw = np.broadcast_to(ipcw[:, np.newaxis], (n_samples, n_times))

    # sort each time point (columns) by risk score (descending)
    o = np.argsort(-estimate, axis=0)
    test_time = np.take_along_axis(test_time, o, axis=0)
    test_event = np.take_along_axis(test_event, o, axis=0)
    estimate = np.take_along_axis(estimate, o, axis=0)
    ipcw = np.take_along_axis(ipcw, o, axis=0)

    is_case = (test_time <= times_2d) & test_event
    is_control = test_time > times_2d
    n_controls = is_control.sum(axis=0)

    # prepend row of infinity values
    estimate_diff = np.concatenate((np.broadcast_to(np.infty, (1, n_times)), estimate))
    is_tied = np.absolute(np.diff(estimate_diff, axis=0)) <= tied_tol

    cumsum_tp = np.cumsum(is_case * ipcw, axis=0)
    cumsum_fp = np.cumsum(is_control, axis=0)
    true_pos = cumsum_tp / cumsum_tp[-1]
    false_pos = cumsum_fp / n_controls

    scores = np.empty(n_times, dtype=float)
    it = np.nditer((true_pos, false_pos, is_tied), order="F", flags=["external_loop"])
    with it:
        for i, (tp, fp, mask) in enumerate(it):
            idx = np.flatnonzero(mask) - 1
            # only keep the last estimate for tied risk scores
            tp_no_ties = np.delete(tp, idx)
            fp_no_ties = np.delete(fp, idx)
            # Add an extra threshold position
            # to make sure that the curve starts at (0, 0)
            tp_no_ties = np.r_[0, tp_no_ties]
            fp_no_ties = np.r_[0, fp_no_ties]
            scores[i] = np.trapz(tp_no_ties, fp_no_ties)

    scores[np.isnan(scores)] = 0
    if n_times == 1:
        mean_auc = scores[0]
    else:
        surv = KaplanMeierFitter()
        surv.fit(survival_test[TIME_NAME], survival_test[CENS_NAME]) 
        s_times = surv.survival_function_at_times(times).to_numpy()
        
        # compute integral of AUC over survival function
        d = -np.diff(np.r_[1.0, s_times])
        integral = (scores * d).sum()
        mean_auc = integral / (1.0 - s_times[-1])

    return mean_auc

"""ESTIMATE FUNCTION"""

def get_survival_func(ddeath, cdeath, bins = None):
    """
    Build Kaplan-Meier Estimate of survival function

    Parameters
    ----------
    ddeath : array-like
        Times of occured events
    cdeath : array-like
        Indicate of occured events (Censoring flag)
    bins : array-like, optional
        Points of surival function. The default is None.

    Returns
    -------
    KaplanMeierFitter or array
        If bins is None return kaplan-meier model, 
                   else return values of SF.

    """
    kmf = KaplanMeierFitter()
    kmf.fit(ddeath, cdeath)
    if not(bins is None):
        return kmf.survival_function_at_times(bins).to_numpy()
    return kmf

def get_hazard_func(ddeath, cdeath, bins = None):
    """
    Build Nelson-Aalen Estimate of Hazard function

    Parameters
    ----------
    ddeath : array-like
        Times of occured events
    cdeath : array-like
        Indicate of occured events (Censoring flag)
    bins : array-like, optional
        Points of hazard function. The default is None.

    Returns
    -------
    NelsonAalenFitter or array
        If bins is None return Nelson-Aalen model, 
                   else return values of HF.

    """
    naf = NelsonAalenFitter()
    naf.fit(ddeath, cdeath)
    if not(bins is None):
        return naf.cumulative_hazard_at_times(bins).to_numpy()
    return naf

# def get_norm_hist(x, b):
#     a = np.histogram(x,bins = b)[0]
#     return a/sum(a)

# def plot_predict_surv(surves, bins, true_time, cens = 1):
#     """
#     RUN:
#         res = ssc.get_score_survival_methods(X_tv, X_tst, new_sign[:-9], bins, True)
#         plot_predict_surv({i:j[0] for i,j in res.items()}, bins, 150.0, 1)
#     """
#     fig, ax = plt.subplots()
    
#     for name, surv in surves.items():
#         ax.plot(bins, surv, 
#                 label = f"{name}, prob:{round(min(surv[np.where(bins < true_time)]),3)}")
#     ax.vlines(true_time, 0, 1, 
#               color = 'k', 
#               linestyles = ('dashed' if cens else 'solid'),
#               linewidth = 2)
#     ax.legend()
#     plt.ylim(-0.1, 1.1)
#     plt.show()

# def print_importance(feature, importance):
#     plt.subplots(figsize=(30, 8))
#     res = {f:i for f,i in zip(feature, importance)}
#     res = dict(sorted(res.items(), key = lambda x: x[1]))
#     print(res)
#     res = dict(sorted(res.items(), key = lambda x: abs(x[1]))[-15:])
#     res = dict(sorted(res.items(), key = lambda x: x[1]))
#     plt.bar(res.keys(), res.values())
#     plt.show()


short = False

CRAID_param_grid = {
    "depth": [10],
    "criterion": ["peto"] if short else ["peto", "tarone-ware", "wilcoxon", "logrank"],
    "min_samples_leaf": [10] if short else [500, 1000, 2000],
    'cut' : [False],
    "woe" : [False], #if short else [True, False], 
    "signif": [0.05] if short else [0.05, 0.1, 0.15],
    "max_features": [1.0],
    "n_jobs" : [40]
}

BSTR_param_grid = {
    "size_sample": [0.7], 
    "n_estimators": [10] if short else [10, 30], 
    "depth": [10],
    "ens_metric_name": ["conc", "ibs"],
    # "woe" : [], 
    "criterion": ["peto"] if short else ["peto", "tarone-ware", "wilcoxon", "logrank"], 
    "min_samples_leaf": [500, 1000, 2000],
    "max_features": [0.7] if short else ["sqrt"],
    "n_jobs" : [40]
}

BOOST_param_grid = {
    "size_sample": [0.5] if short else [0.5],#, 0.7],
    "n_estimators": [15], 
    "ens_metric_name": ["ibs"] if short else ["conc","ibs"],
    "depth": [15],
    "mode_wei": ['exp'] if short else ['square'],#,'exp'],
    # "woe" : [],
    "criterion": ["logrank"] if short else ["peto", "tarone-ware", "wilcoxon", "logrank"],
    "min_samples_leaf": [10] if short else [100], #[500, 1000, 2000],
    "max_features": [0.3], #["sqrt"],
    "aggreg_func": ['wei'] if short else ['wei', 'mean'],
    "n_jobs" : [40]
}

COVID_PARAMS = {
    "TREE": CRAID_param_grid,
    "BSTR": BSTR_param_grid,
    "BOOST": BOOST_param_grid
}


import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
import seaborn as sns
sns.set()
from graphviz import Digraph

from joblib import Parallel, delayed

from .find_split import best_attr_split
from .. import metrics as metr
from .. import constants as cnt

""" Auxiliary functions """

def join_dict(a,b):
    return dict(list(a.items()) + list(b.items()))

""" Класс вершины дерева решений """
class Node(object):
    """
    Node of decision tree.
    Allow to separate data into 2 subnodes (references store in edges) 

    Attributes
    ----------
    df : Pandas DataFrame
        Data of Node
    numb : int
        Number or name of Node
    depth : int
        Distance from root node
    edges : array-like
        Subbranches after separating
    features : list
        Available features
    categ : list
        Names of categorical features
    woe : boolean
        Mode of categorical preparation
    rule : dict
        Allow to define data to node.
        name: str
            query in pandas terms
        attr: str
            feature of separation
        pos_nan: int
            Indicator of nan
    is_leaf : boolean
        True if node don't have subnodes
    verbose : int
        Print best split of node
    info : dict
        Parameters for finding the best split

    Methods
    -------
    check_params : Fill empty parameters and map max_features to int
    find_best_split : Choose best split of node according to parameters
    split : Try to create subnodes by best split
    get_df_node : Return data for node
    set_leaf : Delete subnodes and reset data
    
    predict : Return statistic values of data
    predict_rules : Return full rules from node to leaf
    predict_scheme : Return all possible outcomes for additional features determination
    
    get_figure : Create picture of data (hist, survival function)
    get_rule : Return rule of node
    get_description : Return common values of data (size, depth, death, cens)
    build_viz : Create and fill graphviz digraph
    translate : Replace rules and features by dictionary
    
    """
    __slots__ = ("df", "numb",
                 "depth", "edges", "features", 
                 "categ", "woe", "rule", "is_leaf", "verbose", "info")
    def __init__(self, df,  numb = 1, depth = 0,
                 features = [], categ = [], woe = False,
                 rule = {"name":"","attr":"","pos_nan":0},
                 verbose = 0, **info):
        self.df = df
        self.numb = numb
        self.depth = depth
        self.edges = np.array([])
        self.features = features
        self.categ = categ
        self.woe = woe
        self.rule = rule
        self.is_leaf = True
        self.verbose = verbose
        self.info = info
        self.check_params()
    
    def check_params(self):
        self.info.setdefault("bonf", True)
        self.info.setdefault("n_jobs", 16)
        self.info.setdefault("max_features", 1.0)
        self.info.setdefault("signif", 1.1)
        self.info.setdefault("thres_cont_bin_max", 100)
        if self.info["max_features"] == "sqrt":
            self.info["max_features"] = int(np.trunc(np.sqrt(len(self.features))+0.5))
        elif isinstance(self.info["max_features"],float):
            self.info["max_features"] = int(self.info["max_features"]*len(self.features))
        return
        
    """ GROUP FUNCTIONS: CREATE LEAFS """
    
    def find_best_split(self):
        numb_feats = self.info["max_features"]
        numb_feats = np.clip(numb_feats, 1, len(self.features))
        n_jobs = min(numb_feats,self.info["n_jobs"])
        selected_feats = np.random.choice(self.features, size=numb_feats, replace=False)
        
        args = np.array([])
        for feat in selected_feats:
            t = self.info.copy()
            t["type_attr"] = "woe" if self.woe else "categ" if feat in self.categ else "cont"
            t["arr"] = self.df.loc[:,[feat, cnt.CENS_NAME, cnt.TIME_NAME]].to_numpy().T
            args = np.append(args, t)
        with Parallel(n_jobs=n_jobs, verbose=0, batch_size=10) as parallel:
            ml = parallel(delayed(best_attr_split)(**a) for a in args)

        attrs = {f: ml[ind] for ind,f in enumerate(selected_feats)}
        attr = min(attrs,key=lambda x:attrs[x]["p_value"])
        
        if attrs[attr]["sign_split"] > 0 and self.info["bonf"]:
            attrs[attr]["p_value"] = attrs[attr]["p_value"] / attrs[attr]["sign_split"]
        return (attr, attrs[attr])
        
    def split(self):
        attr, best_split = self.find_best_split()
        # В лучшем признаке не было ни одного значимого разбиения
        if best_split["sign_split"] == 0:
            if self.verbose > 0:
                print(f'Конец ветви, незначащее p-value: {best_split["p_value"]}')
            return (attr, best_split)
        
        if self.verbose > 0:
            print('='*6, best_split["p_value"], attr)
        leaf_ind = 0
        for v,p_n in zip(best_split["values"], best_split["pos_nan"]):
            query = attr + v
            if p_n == 1:
                query = "(" + attr + v + ") or (" + attr + " != " + attr + ")"
            d_node = self.df.query(query).copy()
            N = Node(df = d_node,
                 features = self.features, categ = self.categ,
                 depth = self.depth+1, numb = self.numb*2+leaf_ind,
                 rule = {"name":attr + v, "attr":attr, "pos_nan":p_n},
                 verbose = self.verbose, **self.info)
            self.edges = np.append(self.edges, N)
            leaf_ind+=1
        self.is_leaf = False
        self.df = None
        
    """ GROUP FUNCTIONS: CLEAR AND DEL """
    def get_df_node(self):
        if self.is_leaf:
            return self.df
        return pd.concat([edge.get_df_node() for edge in self.edges])
    
    def set_leaf(self):
        if self.is_leaf:
            return
        self.df = self.get_df_node()
        del self.edges
        self.edges = np.array([])
        self.is_leaf = True
        
    """ GROUP FUNCTIONS: PREDICT """
        
    def predict(self, X, target, name_tg = "res", end_list = []):
        """
        Return statistic values of data

        Parameters
        ----------
        X : Pandas dataframe
            Contain input features of events.
        target : str or function
            Column name, mode or aggregate function of leaf sample.
            Column name : must be in dataset.columns
                Return mean of feature
            Mode :
                "depth" return leafs depth
                "num_node" return leafs numb (names)
            Function : 
                Function of leaf sample (format like target(X_node = dataset))
        name_tg : str, optional
            Name of return column. The default is "res".
        end_list : list, optional
            Numbers of end node (instead leaf). The default is [].

        Returns
        -------
        X[name_tg] : array-like 
            Values by target

        """
        if (self.numb in end_list) or self.is_leaf:
            dataset = self.get_df_node()
            if type(target) != str:
                pred = target(X_node = dataset)
                X.loc[:,name_tg] = X[name_tg].apply(lambda x: pred)
            elif target == "depth":
                X.loc[:,name_tg] = self.depth
            elif target == "num_node":
                X.loc[:,name_tg] = self.numb
            elif target in dataset.columns:
                X.loc[:,name_tg] = np.mean(dataset[target])
        else:
            attr = self.edges[0].rule['attr']
            if attr not in X.columns:
                X.loc[:,attr] = np.nan
            ind_nan = X.index
            for edge in self.edges:
                ind_nan = ind_nan.difference(X.query(edge.rule["name"]).index)
                
            for edge in self.edges:
                ind = X.query(edge.rule["name"]).index
                if edge.rule["pos_nan"] == 1:
                    ind = ind.append(ind_nan)
                if len(ind) > 0:
                    X.loc[ind, name_tg] = edge.predict(X.loc[ind, :], target, name_tg, end_list)
        return X[name_tg]
    
    def predict_rules(self, X, name_tg = "res"):
        if self.is_leaf:
            X.loc[:,name_tg] = self.get_rule()
        else:
            attr = self.edges[0].rule['attr']
            if attr not in X.columns:
                X.loc[:,attr] = np.nan
            ind_nan = X.query(attr + "!=" + attr).index
            for edge in self.edges:
                ind = X.query(edge.rule["name"]).index
                if edge.rule["pos_nan"] == 1:
                    ind = ind.append(ind_nan)
                if len(ind) > 0:
                    X.loc[ind, name_tg] = edge.predict_rules(X.loc[ind, :], name_tg)
                    if len(self.rule["name"]) > 0:
                        X.loc[ind, name_tg] = self.get_rule() + '&' + X.loc[ind, name_tg]
        return X[name_tg]
    
    def get_values_column(self, columns):
        return [0,1]
    
    def predict_scheme(self, X, scheme_feat):
        """
        Return all possible outcomes for additional features determination

        Parameters
        ----------
        X : Pandas dataframe
            Contain input features of events.
        scheme_feat : list
            Features with missing values (relatively).
            If feature in list was used in node, 
                then node consider all possible replaces for value in branches
                Thus, method allow to return all outcomes for different values

        Returns
        -------
        X['res'] : array-like
            For each observation contain dict
                key : rule of overdefined feature (; is separator)
                value : list of sample values
                    censoring flag, time, all values for scheme_feat
                    
        """
        def scheme_output_format(r):
            to_array = lambda col: np.array(self.df.get(col))
            return {r['store_str']:
                    [to_array(cnt.CENS_NAME), 
                     to_array(cnt.TIME_NAME), 
                     #to_array(self.info["sum"]), ## TODO FOR SCHEME'S SUM
                     {sch: to_array(sch) for sch in scheme_feat}]}
            
        def join_scheme_leafs(X_sub, ind_nan = []):
            for edge in self.edges:
                ind = X_sub.query(edge.rule['name']).index
                if len(ind_nan) > 0:
                    if edge.rule["pos_nan"] == 1:
                        ind = ind.append(ind_nan)
                if len(ind) > 0:
                    X_sub.loc[ind, 'tmp'] = edge.predict_scheme(X_sub.loc[ind,:], scheme_feat)
                    X_sub.loc[ind, 'res'] = X_sub.loc[ind,:].apply(lambda r: join_dict(r['res'],r['tmp']), axis = 1)
            return X_sub['res']
            
        if self.is_leaf:
            return X.apply(scheme_output_format, axis = 1)
        attr = self.edges[0].rule['attr']
        if attr not in X.columns:
            X.loc[:,attr] = np.nan
        ind_nan = X.query(attr + "!=" + attr).index
        ind_has = X.index.difference(ind_nan)
        if attr not in scheme_feat:
            X.loc[:,'res'] = join_scheme_leafs(X, ind_nan)
        else:
            if len(ind_has) > 0:
                X.loc[ind_has, 'res'] = join_scheme_leafs(X.loc[ind_has,:])
            if len(ind_nan) > 0:
                pred_store = X.loc[ind_nan, 'store_str'].copy()
                for val in self.get_values_column(attr):
                    X.loc[ind_nan, attr] = val
                    X.loc[ind_nan, 'store_str'] = pred_store + attr + '==' + str(val) + ';'
                    X.loc[ind_nan, 'res'] = join_scheme_leafs(X.loc[ind_nan,:])
        return X['res']
    
    """ GROUP FUNCTIONS: VISUALIZATION """
    
    def get_figure(self, mode = "hist", target = None, save_path = ""):
        if len(save_path) > 0:
            plt.ioff()
        fig, ax = plt.subplots(figsize=(8, 6))
        local_df = self.get_df_node()
        if mode == "hist":
            local_df[target].hist(bins = 25)
            ax.set_xlim([0, np.max(local_df[target])])
        elif mode == "surv":
            kmf = metr.get_survival_func(local_df[cnt.TIME_NAME], local_df[cnt.CENS_NAME])
            ax.set_xlim([0, np.max(local_df[cnt.TIME_NAME])])
            ax.set_ylim([0, 1])
            plt.xticks(range(0, np.max(local_df[cnt.TIME_NAME])+1, 1000))
            kmf.plot_survival_function(legend = False, fontsize=25)
            # ax.set_xlabel('Время', fontsize=25)
            # ax.set_ylabel('Вероятность выживания', fontsize=25)
            ax.set_xlabel('Time', fontsize=25)
            ax.set_ylabel('Survival probability', fontsize=25)#plt.xlabel('Timeline', fontsize=0)
        if len(save_path) > 0:
            plt.savefig(save_path)
        else:
            plt.show()
    
    def get_rule(self):
        if not(self.rule["pos_nan"]):
            return f'({self.rule["name"]})'
        #return f'(({self.rule["name"]})|({self.rule["attr"]} != {self.rule["attr"]}))'
        return f'(({self.rule["name"]})| не указано)'
    
    def get_description(self, full = False):
        s = "" #if not(self.rule["pos_nan"]) else " or " + self.rule["attr"] + " == NaN"
        d = self.get_df_node()
        m_cens = round(d[cnt.CENS_NAME].mean(),2)
        m_time = round(d[cnt.TIME_NAME].mean(),2)
        if full:
            label = "\n".join([self.rule["name"] + s,
                               "size = %s" % (d.shape[0]),
                               "cens/size = %s" % (m_cens),
                               "depth = %s" % (self.depth),
                               "death = %s" % (m_time)])
        else:
            label = self.rule["name"] + s 
        return label
        
    def build_viz(self, dot = None, path_dir = "", depth = None, **args):
        if dot is None:
            dot = Digraph()
        img_path = path_dir + str(self.numb) + '.png'
        self.get_figure(save_path = img_path, **args)
        dot.node(str(self.numb), 
                      label = self.get_description(), 
                      image = img_path, fontsize='30') # fontsize='16'
        if not(depth is None):
            if depth < self.depth:
                return dot
        for ind_e in range(self.edges.shape[0]):
            dot = self.edges[ind_e].build_viz(dot, path_dir, **args)
            dot.edge(str(self.numb), str(self.edges[ind_e].numb))
        return dot
         
    def translate(self, describe):
        if self.is_leaf:
            self.df = self.df.rename(describe, axis = 1)
        self.features = [describe.get(f,f) for f in self.features]
        self.categ = [describe.get(c,c) for c in self.categ]
        self.rule["name"] = describe.get(self.rule["name"],self.rule["name"])
        for edge in self.edges:
            edge.translate(describe)



import pandas as pd
import numpy as np

import matplotlib.pyplot as plt
import seaborn as sns
sns.set()

from .. import metrics as metr
from .. import criteria as scrit

def get_sch_top_descr(df, feats, thres = 5, top = 3):
    df['scheme'] = df.loc[:,feats].apply(lambda r: '*'.join(sorted([ni for ni in feats if r[ni] > 0])), axis = 1)
    
    d = pd.Series(df['scheme'].value_counts()).to_dict()
    #thres = min(thres, sorted(d.values())[-3:][0])
    d = {i:j for i,j in d.items() if j >= thres}
    top_3 = sorted(d, key = lambda x: d[x])[-top:][::-1]
    ret = {}
    for t in top_3:
        tmp = df[df['scheme'] == t]#.query(t + '==1')
        ret[t] = [tmp.shape[0],round(tmp['dDeath'].mean(),3),
                               round(tmp['cDeath'].mean(),3)#, TODO
                               # round(tmp['SUM_M'].mean(),3)
                               ]
    return ret

def get_sch_top_all_descr(df, feats, thres = 5, top = 3):
    df['scheme'] = df.loc[:,feats].apply(lambda r: '*'.join(sorted([ni for ni in feats if r[ni] > 0])), axis = 1)
    
    d = pd.Series(df['scheme'].value_counts()).to_dict()
    #thres = min(thres, sorted(d.values())[-3:][0])
    d = {i:j for i,j in d.items() if j >= thres}
    top_3 = sorted(d, key = lambda x: d[x])[-top:][::-1]
    ret = {}
    for t in top_3:
        tmp = df[df['scheme'] == t]#.query(t + '==1')
        ret[t] = [tmp.shape[0],tmp['dDeath'],
                                tmp['cDeath'] #, TODO
                               # tmp['SUM_M']
                               ]
    return ret

def prepare_name(k, descript):
    res = ""
    tmp = k.split('*')
    delete_shared = []
    for i, sch in enumerate(tmp):
        if '_'.join(tmp[:i] + tmp[i+1:]).find(sch) == -1:
            delete_shared.append(sch)
            
    res += '&'.join([descript.get(name, name) for name in delete_shared])
    return res

def get_full_descr_sch(dict_sch, descript):
    res = ""
    ind = 0
    for k,v in dict_sch.items():
        ind+=1
        # res += "СХЕМА " + str(ind) + ' (' + '  '.join(['КОЛ-ВО:'+str(v[0]),'ВЕРОЯТНОСТЬ СМЕРТИ:'+str(v[2]),'СРЕДНЯЯ СУММА:'+str(v[3])]) + ')\n'
        res += "СХЕМА " + str(ind) + ' (' + '  '.join(['КОЛ-ВО:'+str(v[0]),'ВЕРОЯТНОСТЬ СМЕРТИ:'+str(v[2])]) + ')\n'
        delete_shared = []
        tmp = k.split('*')
        for i, sch in enumerate(tmp):
            if '_'.join(tmp[:i] + tmp[i+1:]).find(sch) == -1:
                delete_shared.append(sch)
        res += '\n'.join([descript.get(name,name) for name in delete_shared])            
        res += '\n'
    return res

def export_legend(legend, filename):
    fig  = legend.figure
    fig.canvas.draw()
    bbox  = legend.get_window_extent().transformed(fig.dpi_scale_trans.inverted())
    fig.savefig(filename, dpi="figure", bbox_inches=bbox)

  
def visualize_pers_km(pred_dict, output_dir, descript, name = None):
    items = sorted(pred_dict,key=lambda x: np.mean(pred_dict[x][0]))[:5]
    plt.ioff()
    fig, ax = plt.subplots(figsize=(7, 5))
    legends = []
    i = 0
    for i in items:
        prob = np.mean(pred_dict[i][0])
        days = np.mean(pred_dict[i][1])
        # summ = pred_dict[i][2]
        prob_sch = pred_dict[i][2] #3 because SUM_M
        dict_sch = pred_dict[i][3] #4 because SUM_M
        kmf = metr.get_survival_func(pred_dict[i][1], pred_dict[i][0])
        ax = kmf.plot_survival_function(ax=ax, show_censors = True, censor_styles = {"marker":'o', "ms":6})
        if prob < 1.0:
            legends.append('Лечение завершено')
        descr = ""
        descr += 'ВЕРОЯТНОСТЬ СМЕРТИ:{:.5}'.format(str(prob)) + ' '
        # descr += 'СРЕДНЯЯ СУММА:{}'.format(str(summ))
        descr += '\n\n{}'.format(get_full_descr_sch(dict_sch, descript))#'\n'.join([get_full_sch(k) + ':' + '  '.join(['NUM:'+str(v[0]),'PROB:'+str(v[2]),'SUM:'+str(v[3])]) for k,v in dict_sch.items()]))
        legends.append(descr)
    plt.savefig(output_dir + 'full_figure.png')
    legend = ax.legend(legends, bbox_to_anchor = (0,-0.1), loc = 'upper left', ncol = 1, prop = {'size':10})
    export_legend(legend, output_dir + "full_legend.png")
    #plt.show()

def visualize_all_sh_km(pred_dict, output_dir, name = None):
    items = sorted(pred_dict,key=lambda x: np.mean(pred_dict[x][0]))[:5]
    plt.ioff()
    SCH_i = 1
    ind = 0
    legends = []
    fig, ax = plt.subplots(figsize=(7, 5))
    pict_dict = []
    for i in items:
        for sch in pred_dict[i][4].keys():
            ind += 1
            full_sch = pred_dict[i][4][sch]
            N = full_sch[0]
            dDeath = full_sch[1]
            cDeath = full_sch[2]
            # SUM_M = round(np.mean(full_sch[3]),3)
            prob = np.mean(cDeath)
            days = np.mean(dDeath)
            kmf = metr.get_survival_func(dDeath, cDeath)
            #descr = "Схема №" + str(ind)
            descr = "\nКОЛ-ВО:" + str(N) + ' '
            descr += 'ВЕРОЯТНОСТЬ СМЕРТИ:{:.5}'.format(str(prob)) + ' '
            # descr += 'СРЕДНЯЯ СУММА:{}'.format(str(SUM_M))
            descr += '\n' + prepare_name(sch)
            #    if prob < 1.0:
            #    legends.append('Лечение завершено')            
            if name is None:
                ax = kmf.plot_survival_function(ax=ax, censor_styles = {"marker":'o', "ms":6})#, show_censors = True)
                legends.append(descr)
            else:
                pict_dict.append([prob, N, kmf, descr])# pict_dict.append([prob, SUM_M, N, kmf, descr])
    if not(name is None):
        ind_sort = 0
        if name == 'sum':
            ind_sort = 1
        elif name == 'N':
            ind_sort = 2
        s_item = sorted(pict_dict, key = lambda x: x[ind_sort])
        if name == 'N':
            s_item = s_item[::-1]
        for s_i in s_item:
            ax = s_i[3].plot_survival_function(ax=ax, censor_styles = {"marker":'o', "ms":6})#, show_censors = True)
            legends.append(s_i[4])
    if ind > 0:
        add_name = "all_" if name is None else name + '_'
        plt.savefig(output_dir + add_name + "figure.png")
        legend = ax.legend(legends, bbox_to_anchor = (0,-0.1), loc = "upper left", ncol = 1, prop = {'size':10})
        export_legend(legend, output_dir + add_name + "legend.png")
        SCH_i += 1
    else:
        raise ValueError("Не выделены значимые схемы лечения")#"Не выделены значимые схемы лечения")
    #plt.show()
    
def visualize_descr_sh_km(pred_dict, 
                          output_dir, 
                          descript, name = None):
    items = sorted(pred_dict,key=lambda x: np.mean(pred_dict[x][0]))[:5]
    plt.ioff()
    ind = 0
    legends = []
    fig, ax = plt.subplots(figsize=(7, 5))
    pict_dict = []
    for i in items:
        for sch in pred_dict[i][4].keys():
            ind += 1
            full_sch = pred_dict[i][4][sch]
            N = full_sch[0]
            dDeath = full_sch[1]
            cDeath = full_sch[2]
            SUM_M = round(np.mean(full_sch[3]),3)
            prob = round(np.mean(cDeath),3)
            days = np.mean(dDeath)
            kmf = metr.get_survival_func(dDeath, cDeath)
            #descr = "Схема №" + str(ind)
            if name == "inter":
                ax = kmf.plot_survival_function(ax=ax, censor_styles = {"marker":'o', "ms":6})#, show_censors = True)
            elif name == "point": #ci_show 
                ax = kmf.plot_survival_function(ax=ax, censor_styles = {"marker":'o', "ms":6}, show_censors = True, ci_show = False)
                if prob < 1.0:
                    legends.append('Лечение завершено')
            elif name == "p&i":
                ax = kmf.plot_survival_function(ax=ax, censor_styles = {"marker":'o', "ms":6}, show_censors = True, ci_show = True)
                if prob < 1.0:
                    legends.append('Лечение завершено')
            elif name == "without":
                ax = kmf.plot_survival_function(ax=ax, censor_styles = {"marker":'o', "ms":6}, ci_show = False)
            name_plot = prepare_name(sch, descript)
            if name == "inter":
                pict_dict.append([ind, name_plot, prob, SUM_M, N])
            legends.append("СХЕМА №" + str(ind) + '\n' + name_plot.replace("&","\n"))
    if ind > 0:
        add_name = "all_" if name is None else name + '_'
        plt.savefig(output_dir + add_name + "figure.png")
        legend = ax.legend(legends, bbox_to_anchor = (0,-0.1), loc = "upper left", ncol = 1, prop = {'size':10})
        export_legend(legend, output_dir + "shared_legend.png")
        if name == "inter":
            descr = pd.DataFrame(pict_dict,columns = ['Номер','Описание','Вероятность','Стоимость','Количество'])
            descr.to_csv(output_dir + "SchDescr.csv", sep = ";",index = False)
    else:
        raise ValueError("Не выделены значимые схемы лечения")
    #plt.show()
    
def check_correct_list(pred_dict, join = False):
    base = dict()
    for i in pred_dict.keys():
        df = pd.DataFrame(pred_dict[i][2])
        # df = pd.DataFrame(pred_dict[i][3]) #TODO
        df['cDeath'] = pred_dict[i][0]
        df['dDeath'] = pred_dict[i][1]
        # df['SUM_M'] = pred_dict[i][2] #TODO
        base[i] = df
    if not join:
        return base
    while len(base.keys()) > 1:
        p_value_d = dict()
        for i1,l1 in enumerate(base.keys()):
            for i2,l2 in enumerate(base.keys()):
                if i2 > i1:
                    p_val = scrit.logrank_fast(base[l1]['dDeath'],base[l2]['dDeath'],
                                         base[l1]['cDeath'],base[l2]['cDeath'])
                    p_value_d[l1+'#'+l2] = p_val
        p_val_s = sorted(p_value_d, key = lambda x:p_value_d[x])
        print('Максимальное P-value:', p_value_d[p_val_s[-1]])
        if p_value_d[p_val_s[-1]] < 0.05:
            break
        f_l, s_l = p_val_s[-1].split('#')
        print('Цепочки схем:', f_l, s_l)
        print('Заменяются на:', f_l + '|' + s_l)
        base[f_l + '|' + s_l] = pd.concat([base[f_l].copy(),base[s_l].copy()])
        del base[f_l]
        del base[s_l]
    return base
        

# descript = cnt.mapp_sch
def pred_pers_km_count(pred_dict, output_dir, 
                       join = True, 
                       min_size = 20,
                       top_count = 100,
                       func_vis = visualize_pers_km, 
                       func_get_sch = get_sch_top_descr,
                       name = None,
                       descript = {}):
    res = dict()
    base = check_correct_list(pred_dict, join)
    for i in base.keys():
        df = base[i]
        scheme_feat = list(set(df.columns) - {'cDeath','dDeath'}) #TODO: ,'SUM_M'})
        tmp = [df['cDeath'],
               df['dDeath'],
               # round(np.mean(df['SUM_M']),2), #TODO
               {sch: df[sch].mean() for sch in scheme_feat},
               func_get_sch(df, scheme_feat, thres = min_size, top = top_count)
              ]
        if tmp[-1] != {}:
            res[i] = tmp
    func_vis(res, output_dir = output_dir, name = name, descript = descript)


import pandas as pd
import numpy as np

# from .. import metrics as metr
from ..tree import CRAID
from .. import constants as cnt
from .base_ensemble import BaseEnsemble

class BootstrapCRAID(BaseEnsemble):
    """
    Bootstrap aggregation (Bagging) ensemble of survival decision tree.
    On each iteration probabilities of observations change by scheme.

    Attributes
    ----------
    kwargs : dict
        Parameters for building base ensemble (look at BaseEnsemble)

    Methods
    -------
    
    fit : build ensemble with X, y data
    
    """
    def __init__(self, **kwargs):
        self.name = "BootstrapCRAID"
        super().__init__(**kwargs)
    
    def fit(self, X, y):
        self.features = X.columns
        X = X.reset_index(drop=True)
        X[cnt.CENS_NAME] = y[cnt.CENS_NAME].astype(np.int32)
        X[cnt.TIME_NAME] = y[cnt.TIME_NAME].astype(np.int32)
        
        self.X_train = X
        self.y_train = y
        self.update_params()
        
        for i in range(self.n_estimators):
            x_sub = self.X_train.sample(n = self.size_sample, replace=self.bootstrap, random_state=i)
            x_oob = self.X_train.loc[self.X_train.index.difference(x_sub.index),:]
            
            x_sub = x_sub.reset_index(drop=True)
            X_sub_tr, y_sub_cr = cnt.pd_to_xy(x_sub)
            
            model = CRAID(features = self.features, random_state = i, **self.tree_kwargs)
            model.fit(X_sub_tr, y_sub_cr)         

            self.add_model(model, x_oob)
            self.ens_metr[i] = self.score_oob()
            if not (self.tolerance) and i > 0:
                print(f"METRIC: {self.ens_metr[i-1]} -> +1 model METRIC: {self.ens_metr[i]}")
                if self.descend_metr:
                    stop = self.ens_metr[i-1] < self.ens_metr[i]
                else:
                    stop = self.ens_metr[i-1] > self.ens_metr[i]
                if stop:
                    self.select_model(0,len(self.models)-1)
                    break
        
        if self.tolerance:
            self.tolerance_find_best()
        print('fitted:',len(self.models),'models.')


import pandas as pd
import numpy as np

from .. import metrics as metr
from ..tree import CRAID
from .. import constants as cnt
from .base_ensemble import BaseEnsemble

def loss_func(var, mode = 'linear'):
    D = var.max()
    if mode == 'linear':
        return var/D
    elif mode == 'square':
        return (var/D)**2
    elif mode == 'exp':
        return 1.0 - np.exp(-var/D)
    return None

def count_weight(losses, mode = 'linear'):
    li = loss_func(losses, mode = mode)
    l_mean = li.mean() #(li * pred_wei).sum()/pred_wei.sum()
    betta = l_mean/(1.0 - l_mean)
    new_wei = betta ** (1 - li)
    return new_wei, betta


class BoostingCRAID(BaseEnsemble):
    """
    Adaptive boosting ensemble of survival decision tree.
    On each iteration probabilities of observations change by scheme.

    Attributes
    ----------
    mode_wei : str
        Type of weighting scheme
    kwargs : dict
        Parameters for building base ensemble (look at BaseEnsemble)
    
    weights : array-like 
        Current weights for observation
    bettas : list
        Confidence coeff for each base model 
    l_weights : list
        List of weights array for each observation

    Methods
    -------
    
    fit : build ensemble with X, y data
    count_model_weights : count weights according ibs and mode scheme
    add_model : add model, weights and bettas
        
    References
    ----------
    .. [1] Drucker H. Improving regressors using boosting techniques 
            //ICML. – 1997. – Т. 97. – С. 107-115.
    
    """
    def __init__(self, mode_wei = "linear", **kwargs):
        self.name = "BoostingCRAID"
        self.mode_wei = mode_wei
        super().__init__(**kwargs)
        
    def fit(self, X, y):
        self.features = X.columns
        X = X.reset_index(drop=True)
        X[cnt.CENS_NAME] = y[cnt.CENS_NAME].astype(np.int32)
        X[cnt.TIME_NAME] = y[cnt.TIME_NAME].astype(np.int32)
        
        self.X_train = X
        self.X_train["ind_start"] = self.X_train.index
        self.y_train = y
        
        self.weights = np.ones(self.X_train.shape[0], dtype = float)
        self.bettas = []
        self.l_weights = []
        self.update_params()
        
        for i in range(self.n_estimators):
            x_sub = self.X_train.sample(n = self.size_sample, weights = self.weights, 
                                        replace=self.bootstrap, random_state=i)
            x_oob = self.X_train.loc[self.X_train.index.difference(x_sub.index),:]
            
            x_sub = x_sub.reset_index(drop=True)
            X_sub_tr, y_sub_tr = cnt.pd_to_xy(x_sub)
            # X_sub_tr = X_sub_tr.drop('ind_start', axis = 1)
            
            model = CRAID(features = self.features, random_state = i, **self.tree_kwargs)
            model.fit(X_sub_tr, y_sub_tr)
            
            wei_i, betta_i = self.count_model_weights(model, X_sub_tr, y_sub_tr)
            self.add_model(model, x_oob, wei_i, betta_i)
            self.update_weight(x_sub['ind_start'], wei_i)
            
            # print(betta_i)
            self.ens_metr[i] = self.score_oob()
            
            if not (self.tolerance) and i > 0:
                print(f"METRIC: {self.ens_metr[i-1]} -> +1 model METRIC: {self.ens_metr[i]}")
                if self.descend_metr:
                    stop = self.ens_metr[i-1] < self.ens_metr[i]
                else:
                    stop = self.ens_metr[i-1] > self.ens_metr[i]
                if stop:
                    self.select_model(0,len(self.models)-1)
                    break
        
        if self.tolerance:
            self.tolerance_find_best()
        print('fitted:',len(self.models),'models.')
    
    # def count_model_weights(self, model):
    #     pred_surv = model.predict_at_times(self.X_train, bins = self.bins, mode = "surv")
    #     losses = metr.ibs(self.y_train, self.y_train, pred_surv, self.bins, axis = 0)
    #     wei, betta = count_weight(losses, mode = self.mode_wei)
    #     return wei, betta
    
    # def update_weight(self, wei_i):
    #     self.weights[self.X_train.index] *= wei_i
    
    def count_model_weights(self, model, X_sub, y_sub):
        pred_surv = model.predict_at_times(X_sub, bins = self.bins, mode = "surv")
        losses = metr.ibs(self.y_train, y_sub, pred_surv, self.bins, axis = 0)
        wei, betta = count_weight(losses, mode = self.mode_wei)
        return wei, betta
    
    def update_weight(self, index, wei_i):
        self.weights[index] *= wei_i
    
    def add_model(self, model, x_oob, wei_i, betta_i):
        self.bettas.append(betta_i)
        self.l_weights.append(wei_i)
        super().add_model(model, x_oob)
    
    def select_model(self, start, end):
        self.bettas = self.bettas[start:end]
        self.l_weights = self.l_weights[start:end]
        super().select_model(start, end)
    
    def get_aggreg(self, x):
        """
        Overload function of response aggregation.
        Have median, weight (weighted sum with bettas) and mean mode.

        Parameters
        ----------
        x : Pandas dataframe
            Contain input features of events.

        Returns
        -------
        float

        """
        if self.aggreg_func == 'median':
            return np.median(x, axis = 0)
        elif self.aggreg_func == 'wei':
            inv_wei = -1*np.log(self.bettas)
            wei = inv_wei/sum(inv_wei)
            return np.sum((x.T*wei).T, axis = 0)
        return np.mean(x, axis = 0)


short = False

CRAID_param_grid = {
    "depth": [10, 15],
    "criterion": ["peto"] if short else ["peto", "tarone-ware", "wilcoxon", "logrank"],
    "min_samples_leaf": [5] if short else [5],
    'cut' : [True, False],
    "woe" : [False], #if short else [True, False], 
    "signif": [0.05] if short else [0.05, 0.1],
    "max_features": [1.0]
}

BSTR_param_grid = {
    "size_sample": [0.7], 
    "n_estimators": [10] if short else [10, 30],
    "depth": [15],
    "ens_metric_name": ["roc"] if short else ["conc", "ibs", "roc"],
    # "woe" : [], 
    "criterion": ["peto"] if short else ["peto", "tarone-ware", "wilcoxon", "logrank"], 
    "min_samples_leaf": [1, 5, 10],
    "max_features": [0.3] if short else [0.3, "sqrt"]
}

# BOOST_param_grid = {
#     "aggreg_func": ['wei'] if short else ['wei', 'mean'],
#     "criterion": ["logrank"] if short else ["peto", "tarone-ware", "wilcoxon", "logrank"],
#     "depth": [15],
#     "ens_metric_name": ["ibs"] if short else ["conc","ibs"],
#     "max_features": [0.3] if short else [0.3], #"sqrt"],
#     "min_samples_leaf": [1] if short else [1, 5, 15],
#     "mode_wei": ['square', 'exp'] if short else ['square'],#'exp'],
#     "n_estimators": [15] if short else [10, 15, 25],
#     "size_sample": [0.5] if short else [0.5, 0.7],
#     # "woe" : [],
# }

BOOST_param_grid = {
    "aggreg_func": ['wei'] if short else ['wei', 'mean'],
    "criterion": ["logrank"] if short else ["peto", "tarone-ware", "wilcoxon", "logrank"],
    "depth": [15, 25],
    "ens_metric_name": ["ibs"] if short else ["roc"],
    "max_features": [0.3] if short else ["sqrt"],
    "min_samples_leaf": [1] if short else [1, 5, 15],
    "mode_wei": ['square', 'exp'] if short else ['square', 'exp'],
    "n_estimators": [15] if short else [10, 15, 25],
    "size_sample": [0.5] if short else [0.5, 0.7],
    # "woe" : [],
}

PBC_PARAMS = {
    "TREE": CRAID_param_grid,
    "BSTR": BSTR_param_grid,
    "BOOST": BOOST_param_grid
}


import numpy as np
import pandas as pd

from scipy import stats
from lifelines import CoxPHFitter
# from lifelines.statistics import logrank_test
# import fastlogranktest as flr
from numba import njit, jit

### Interface functions
def logrank(durations_A, durations_B, event_observed_A=None, event_observed_B=None):
    return weight_lr_fast(durations_A, durations_B, event_observed_A, event_observed_B)

def wilcoxon(durations_A, durations_B, event_observed_A=None, event_observed_B=None):
    return weight_lr_fast(durations_A, durations_B, event_observed_A, event_observed_B, weightings="wilcoxon")

def peto(durations_A, durations_B, event_observed_A=None, event_observed_B=None):
    return weight_lr_fast(durations_A, durations_B, event_observed_A, event_observed_B, weightings="peto")

def tarone_ware(durations_A, durations_B, event_observed_A=None, event_observed_B=None):
    return weight_lr_fast(durations_A, durations_B, event_observed_A, event_observed_B, weightings="tarone-ware")

CRITERIA_DICT = {
    "logrank": logrank,
    "wilcoxon": wilcoxon,
    "peto": peto,
    "tarone-ware": tarone_ware
}
""" dict: Available criteria in library and its realization """

def logrank_self(durations_A, durations_B, event_observed_A=None, event_observed_B=None) -> float:
    # 27.1 ms ± 1.26 ms per loop (mean ± std. dev. of 7 runs, 10 loops each)
    try:
        event_times_A, event_times_B = (np.array(durations_A), np.array(durations_B))
        time = np.r_[event_times_A, event_times_B]
        group_indicator = np.r_[np.zeros(event_times_A.shape[0], dtype=int), np.ones(event_times_B.shape[0], dtype=int)]
        event = np.r_[event_observed_A, event_observed_B]
        n_samples = time.shape[0]
        groups, group_counts = np.unique(group_indicator, return_counts=True)
        n_groups = groups.shape[0]
        if n_groups == 1:
            raise ValueError("At least two groups must be specified, "
                             "but only one was provided.")

        # sort descending
        o = np.argsort(-time, kind="mergesort")
        x = group_indicator[o]
        event = event[o]
        time = time[o]

        at_risk = np.zeros(n_groups, dtype=np.int_)
        observed = np.zeros(n_groups, dtype=np.int_)
        expected = np.zeros(n_groups, dtype=np.float_)
        covar = np.zeros((n_groups, n_groups), dtype=np.float_)
        k = 0
        while k < n_samples:
            ti = time[k]
            total_events = 0
            while k < n_samples and ti == time[k]:
                idx = np.searchsorted(groups, x[k])
                if event[k]:
                    observed[idx] += 1
                    total_events += 1
                at_risk[idx] += 1
                k += 1

            if total_events != 0:
                total_at_risk = k
                expected += at_risk * (total_events / total_at_risk)
                if total_at_risk > 1:
                    multiplier = total_events * (total_at_risk - total_events) / (total_at_risk * (total_at_risk - 1))
                    for g1 in range(n_groups):
                        temp = at_risk[g1] * multiplier
                        covar[g1, g1] += temp
                        for g2 in range(n_groups):
                            covar[g1, g2] -= temp * at_risk[g2] / total_at_risk

        df = n_groups - 1
        zz = observed[:df] - expected[:df]
        chisq = np.linalg.solve(covar[:df, :df], zz).dot(zz)
        pval = stats.chi2.sf(chisq, df)
    except:
        pval = 1.0
    return pval
    
def cox(durations_A, durations_B, event_observed_A=None, event_observed_B=None) -> float:
    dfA = pd.DataFrame({'E': event_observed_A, 'T': durations_A, 'group': 0})
    dfB = pd.DataFrame({'E': event_observed_B, 'T': durations_B, 'group': 1})
    df_ = pd.concat([dfA, dfB])

    cph = CoxPHFitter().fit(df_, 'T', 'E')
    return cph.log_likelihood_ratio_test().p_value

### 06/02/2022 Numpy+Numba 
### accelerate 2.8 according to library criteria
@njit
def coeffs_t_j(dur_A, dur_B, cens_A, cens_B, t_j, weightings):
    N_1_j = (dur_A >= t_j).sum()#np.where(dur_A >= t_j,1,0).sum()
    N_2_j = (dur_B >= t_j).sum()#np.where(dur_B >= t_j,1,0).sum()
    if N_1_j == 0 or N_2_j == 0:
        return 0, 0, 0
    O_1_j = ((dur_A == t_j) * cens_A).sum() #np.where(dur_A == t_j, cens_A,0).sum()
    O_2_j = ((dur_B == t_j) * cens_B).sum()#np.where(dur_B == t_j, cens_B,0).sum()
    
    N_j = N_1_j + N_2_j
    O_j = O_1_j + O_2_j
    E_1_j = N_1_j*O_j/N_j
    w_j = 1
    if weightings == "wilcoxon":
        w_j = N_j
    elif weightings == "tarone-ware":
        w_j = np.sqrt(N_j)
    elif weightings == "peto":
        w_j = (1.0 - float(O_j)/(N_j+1))
    
    num = O_1_j - E_1_j
    denom = E_1_j*(N_j - O_j) * N_2_j/(N_j*(N_j - 1))
    return w_j, num, denom

@njit
def lr_statistic(dur_A, dur_B, cens_A, cens_B, times, weightings):
    res = np.zeros((times.shape[0],3), dtype=np.float32)
    for j, t_j in enumerate(times):
        res[j] = coeffs_t_j(dur_A, dur_B, cens_A, cens_B, t_j, weightings)
    
    if weightings == "peto":
        res[:, 0] = np.cumprod(res[:, 0])
    logrank = np.power((res[:, 0]*res[:, 1]).sum(),2) / ((res[:, 0]*res[:, 0]*res[:, 2]).sum())
    return logrank

def weight_lr_fast(dur_A, dur_B, cens_A = None, cens_B = None, weightings = ""):
    """
    Count weighted log-rank criteria

    Parameters
    ----------
    dur_A : array-like 
        Time of occured events from first sample.
    dur_B : array-like 
        Time of occured events from second sample.
    cens_A : array-like, optional
        Indicate of occured events from first sample. 
        The default is None (all events occured).
    cens_B : array-like, optional
        Indicate of occured events from second sample. 
        The default is None (all events occured).
    weightings : str, optional
        Weights of criteria. The default is "" (log-rank).
        Log-rank :math:'w = 1'
        Wilcoxon :math:'w = N_j'
        Tarone-ware :math:'w = \\sqrt(N_j)'
        Peto-peto :math:'w = \\fraq{1 - O_j}{N_j + 1}'

    Returns
    -------
    p-value : float
        Chi2 p-value of weighted log-rank test

    """
    try:
        if cens_A is None:
            cens_A = np.ones(dur_A.shape[0])
        if cens_B is None:
            cens_B = np.ones(dur_B.shape[0])

        #     a1 = np.unique(dur_A)
        #     a2 = np.unique(dur_B)
        #     times = np.unique(np.clip(np.union1d(a1,a2), 0, np.min([a1.max(), a2.max()])))
        times = np.union1d(np.unique(dur_A),np.unique(dur_B))
        logrank = lr_statistic(dur_A, dur_B, cens_A, cens_B, times, weightings)
        pvalue = stats.chi2.sf(logrank, df = 1)
        return pvalue
    except:
        return 1.0


import pandas as pd
import numpy as np
import os
import time
import copy

import matplotlib.pyplot as plt
import seaborn as sns
sns.set()
from graphviz import Digraph
from sklearn.metrics import roc_auc_score
from sklearn.preprocessing import OneHotEncoder
from sksurv.linear_model import CoxPHSurvivalAnalysis

from .. import metrics as metr
from .node import Node
from .. import constants as cnt


def format_to_pandas(X, columns):
    type_df = type(X)
    if type_df.__name__ == "DataFrame":
        return X.loc[:,columns]
    elif type_df.__name__ == "ndarray":
        return pd.DataFrame(X, columns = columns)
    return None

""" Functions of prunning """

def ols(a,b):
    return sum((a - b)**2)

def get_full_copy(tree):
    return copy.deepcopy(tree)

def count_leaf(tree):
    if tree.is_leaf:
        return np.array([tree.numb])
    arr = np.array([])
    for edge in tree.edges:
        arr = np.append(arr, count_leaf(edge))
    return arr.flatten().astype('int')

def count_spanning_leaf(tree):
    if tree.is_leaf:
        return np.array([])
    arr = np.array([])
    for edge in tree.edges:
        arr = np.append(arr, count_spanning_leaf(edge))
    if arr.shape[0] == 0:
        return np.array([tree.numb])
    return arr.flatten().astype('int')

def delete_sub_leaf(tree, leaf_list):
    if tree.numb in leaf_list:
        tree.set_leaf()
    for edge in tree.edges:
        delete_sub_leaf(edge, leaf_list)

def find_best_uncut(tree, X, y, target, mode_f, choose_f):
    span_leaf = count_spanning_leaf(tree)
    # print(span_leaf)
    d = {}
    for el in span_leaf:
        y_pred = tree.predict(X, target = target, end_list = [el])
        d[el] = round(mode_f(y,y_pred),4)
    
    new_leaf, val = choose_f(d.items(), key = lambda x: x[1])
    delete_sub_leaf(tree, [new_leaf])
    return tree, val
    
def cutted_tree(tree_, X, target, mode_f, choose_f, verbose = 0):
    first_digits = lambda x: float(str(x)[:5])
    y = pd.to_numeric(X[target])
    tree = get_full_copy(tree_)
    best_metr = dict()
    best_tree = dict()
    y_pred = tree.predict(X, target = target)
    c = count_leaf(tree).shape[0]
    
    best_metr[c] = mode_f(y, y_pred)
    best_tree[c] = get_full_copy(tree)
    while (tree.edges.shape[0] > 0):
        tree, val = find_best_uncut(tree, X, y, target, mode_f, choose_f)
        c = count_leaf(tree).shape[0]
        best_metr[c] = val
        best_tree[c] = get_full_copy(tree)
    
    best_metric = first_digits(choose_f(best_metr.values()))
    min_leaf = min([k for k,v in best_metr.items() if first_digits(v) == best_metric])
    
    if verbose > 0:
        plt.clf()
        plt.plot(list(best_metr.keys()), list(best_metr.values()),'o')
        # plt.plot(list(best_metr.keys()), list(best_metr.values()),'b')
        plt.xlabel("Количество листов")#("Leafs")
        plt.ylabel(f"Лучшее значение метрики {mode_f.__name__}")# {target}")
        plt.title(f"Обрезка дерева по переменной {target}")
        plt.show()
        print(best_metr)
        print(best_metric, min_leaf)
    
    return best_tree[min_leaf]

class CRAID(object):
    """
    Survival decision tree model.

    Attributes
    ----------
    depth : int
        Maximal depth of nodes
    cut : boolean
        Flag of prunning
    features : list
        Available features
    categ : list
        Names of categorical features
    tree : Node 
        Root node of model
    coxph : CoxPHSurvivalAnalysis
        Model for hazard prediction
    ohenc : OneHotEncoder
        Encoding model from number of Node to indicators
    bins : array-like
        Points of timeline.
    info : dict
        Parameters for building nodes

    Methods
    -------
    
    fit : build decision tree with X, y data (recursive splitting node)
    predict : return values of features, rules or schemes
    predict_at_times : return survival or hazard function
    cut_tree : prunning function
    
    visualize : build graphviz Digraph for each node
    
    """
    def __init__(self, depth = 0,
                 random_state = 123,
                 features = [],
                 categ = [],
                 cut = False,
                 **info):
        self.info = info
        self.cut = cut
        self.remove_files = []
        self.tree = None
        self.depth = depth
        self.features = features
        self.categ = categ
        self.random_state = random_state
        self.name = "CRAID_%s" % (self.random_state)
        self.coxph = None
        self.ohenc = None
        self.bins = []
    
    def fit(self, X, y):
        if len(self.features) == 0:
            self.features = X.columns
        self.bins = cnt.get_bins(time = y[cnt.TIME_NAME])#, cens = y[cnt.CENS_NAME])
        X = X.reset_index(drop=True)
        X_tr = X.copy()
        X_tr[cnt.CENS_NAME] = y[cnt.CENS_NAME].astype(np.int32)
        X_tr[cnt.TIME_NAME] = y[cnt.TIME_NAME].astype(np.int32)
        
        if not("min_samples_leaf" in self.info):
            self.info["min_samples_leaf"] = 0.01*X_tr.shape[0]
        cnt.set_seed(self.random_state)
        
        if self.cut:
            X_val = X_tr.sample(n = int(0.2*X_tr.shape[0]), random_state=self.random_state)
            X_tr = X_tr.loc[X_tr.index.difference(X_val.index),:]
         
        # t_start = time.perf_counter()
        self.tree = Node(X_tr, features = self.features, 
                         categ = self.categ, **self.info)
        stack_nodes = np.array([self.tree], dtype = object)
        while(stack_nodes.shape[0] > 0):
            node = stack_nodes[0]
            stack_nodes = stack_nodes[1:]
            if node.depth >= self.depth:
                continue
            node.split()
            for ind_edge in range(node.edges.shape[0]):
                stack_nodes = np.append(stack_nodes, node.edges[ind_edge])
                
        if self.cut:
            self.cut_tree(X_val, cnt.CENS_NAME, mode_f = roc_auc_score, choose_f = max)
        
        self.coxph = CoxPHSurvivalAnalysis(alpha = 0.1)
        self.ohenc = OneHotEncoder(handle_unknown='ignore')
        pred_node = self.tree.predict(X, target = "num_node").to_numpy().reshape(-1,1)
        ohenc_node = self.ohenc.fit_transform(pred_node).toarray()
        self.coxph.fit(ohenc_node, y)
        
        # print(f"FULL_TIME: {time.perf_counter() - t_start} seconds")
        # print('*'*6, 'End fit.', '*'*6)
        return
    
    def predict(self, X, mode = "target", target = cnt.TIME_NAME, name_out = "res"):
        """
        Return values of features, rules or schemes

        Parameters
        ----------
        X : Pandas dataframe
            Contain input features of events.
        mode : str, optional
            Mode of predicting. The default is "target".
            "target" : return values of feature (in target variable)
            "scheme" : return all possible outcomes for features determination
                       target is list of features with missing values (relatively)
            "rules" : return full rules from node to leaf
        target : str or list, optional
            Aim of predicting. The default is occured time.
        name_out : str, optional
            Name of return column. The default is "res".

        Returns
        -------
        X[name_out] : array-like 
            Values by mode & target

        """
        X = format_to_pandas(X, self.features)
        X.loc[:, name_out] = np.nan
        if mode == "target":
            X.loc[:, name_out] = self.tree.predict(X, target)
        elif mode == "scheme":
            X.loc[:,'store_str'] = ""
            X.loc[:, name_out] = X[name_out].apply(lambda x: dict())
            X.loc[:, name_out] = self.tree.predict_scheme(X, target)
        elif mode == "rules":
            X.loc[:, name_out] = self.tree.predict_rules(X)
        return X[name_out]
    
    def predict_at_times(self, X, bins, mode = "surv"):
        """
        Return survival or hazard function.

        Parameters
        ----------
        X : Pandas dataframe
            Contain input features of events.
        bins : array-like
            Points of timeline.
        mode : str, optional
            Type of function. The default is "surv".
            "surv" : send building function in nodes
            "hazard" : fit CoxPH model on node numbers (input) 
                                          and time/cens (output)
                       predict cumulative HF from model 

        Returns
        -------
        array-like
            Vector of function values in times (bins).

        """
        X = format_to_pandas(X, self.features)
        if mode == "hazard":
            return self.predict_hazard(X, bins)
        def build_at_times(X_node):
            if mode == "surv":
                return metr.get_survival_func(X_node[cnt.TIME_NAME],
                                          X_node[cnt.CENS_NAME], 
                                          bins = bins)
            return metr.get_hazard_func(X_node[cnt.TIME_NAME],
                                          X_node[cnt.CENS_NAME], 
                                          bins = bins)
        X.loc[:,'f_at_times'] = np.nan
        X.loc[:,'f_at_times'] = self.tree.predict(X, build_at_times, name_tg = "f_at_times")
        return np.array(X['f_at_times'].to_list())
        
    def predict_hazard(self, X, bins):
        bins = np.clip(bins, self.bins.min(), self.bins.max())
        pred_node = self.tree.predict(X, target = "num_node").to_numpy().reshape(-1,1)
        ohenc_node = self.ohenc.transform(pred_node).toarray()
        hazards = self.coxph.predict_cumulative_hazard_function(ohenc_node)
        pred_haz = np.array(list(map(lambda x: x(bins), hazards)))
        return pred_haz
    
    def cut_tree(self, X, target, mode_f = roc_auc_score, choose_f = max):
        """
        Method of prunning tree.
        Find best subtree, which reaches best value of metric "mode_f""

        Parameters
        ----------
        X : Pandas dataframe
            Contain input features of events.
        target : str
            Feature name for metric counting.
        mode_f : function, optional
            Metric for selecting. The default is roc_auc_score.
        choose_f : function, optional
            Type of best value (max or min). The default is max.

        """
        self.tree = cutted_tree(self.tree, X, target, mode_f, choose_f)
    
    def visualize(self, path_dir = "", **kwargs):
        tmp_dir = path_dir + self.name + "\\"
        if not os.path.exists(path_dir):
            os.mkdir(path_dir)
        if not os.path.exists(tmp_dir):
            os.mkdir(tmp_dir)
        dot = Digraph(node_attr={'shape': 'none'})
        dot = self.tree.build_viz(dot, path_dir = tmp_dir, **kwargs)
        dot.render(path_dir + self.name + "_", view = False, format = "png")
        for root, dirs, files in os.walk(tmp_dir):
            for file in files:
                os.remove(os.path.join(root, file))
        os.rmdir(tmp_dir)
        
    # def get_score(self, X_v, bins, mode = 'cox'):
    #     print("CRAID", end = " ")
    #     print("IBC: {:.4}".format(self.score(X_v, bins, name = "ibs")), end = " ")
    #     print("CONC: {:.4}".format(self.score(X_v, bins, name = "conc")))
        
    # def score(self, X_v, bins = None, name = "conc"):
    #     X_t = self.tree.get_df_node()
    #     if name == "conc":
    #         pred_time = self.tree.predict(X_v, cnt.TIME_NAME)
    #         return concordance_index(X_v[cnt.TIME_NAME],pred_time)
    #     pred_bins = self.predict_at_times(X_v, bins = bins, mode = "surv")
        
    #     y_train = cnt.get_y(X_t[cnt.CENS_NAME],X_t[cnt.TIME_NAME])
    #     y_true = cnt.get_y(X_v[cnt.CENS_NAME],X_v[cnt.TIME_NAME])
        
    #     return metr.ibs(y_train, y_true, pred_bins, bins)
    
    def translate(self, describe):
        self.features = [describe.get(f,f) for f in self.features]
        self.categ = [describe.get(c,c) for c in self.categ]
        self.tree.translate(describe)


import pandas as pd
import numpy as np

from itertools import chain, combinations
from .. import criteria as scrit 

""" Auxiliary functions """

def power_set_nonover(s):
    """
    Build a list of pairs of nonoverlapping sets
    
    Parameters
    ----------
    s : set
        Contain unique elems.

    Returns
    -------
    a : array
        Contain pairs of nonoverlapping subsets of s elements

    """
    a = np.array(list(chain.from_iterable(combinations(s, r) for r in range(len(s)+1))))
    a = np.vstack([a[:int(a.shape[0]/2)], a[int(a.shape[0]/2):][::-1]]).T
    return a

def transform_woe(x_feat, y):
    """
    Calculate weight of evidence and transform input values.
    Also return numpy array of values mapping

    Parameters
    ----------
    x_feat : array
        Input categorical feature. Nan values ignored
    y : array
        Target binary variable.

    Returns
    -------
    woe_x_feat : array
        Mapped x_feat for woe's values.
    descr_np : array
        Array with 2 dims: source categories, mapping woe's.

    """
    a = np.vstack([x_feat,y]).T
    a = a[a[:,0] == a[:,0]]

    categs = np.unique(a[:,0]).shape[0]
    N_T = y.shape[0]
    N_D = y.sum()
    N_D_ = N_T - y.sum()

    df_woe_iv = pd.crosstab(a[:,0],a[:,1])
    all_0 = df_woe_iv[0].sum()
    all_1 = df_woe_iv[1].sum()

    df_woe_iv["p_bd"] =  (df_woe_iv[1] + 1e-5) / (N_D + 1e-5)
    df_woe_iv["p_bd_"] =  (df_woe_iv[0] + 1e-5) / (N_D_ + 1e-5)
    df_woe_iv["p_b_d"] =  (all_1 - df_woe_iv[1] + 1e-5) / (N_D + 1e-5)
    df_woe_iv["p_b_d_"] =  (all_0 - df_woe_iv[0] + 1e-5) / (N_D_ + 1e-5)

    df_woe_iv["woe_pl"]= np.log(df_woe_iv["p_bd"] / df_woe_iv["p_bd_"])
    df_woe_iv["woe_mn"]= np.log(df_woe_iv["p_b_d"] / df_woe_iv["p_b_d_"])
    features_woe = (df_woe_iv["woe_pl"] - df_woe_iv["woe_mn"]).to_dict()
    descr_np = np.vstack([df_woe_iv.index, (df_woe_iv["woe_pl"] - df_woe_iv["woe_mn"])])
    woe_x_feat = np.vectorize(features_woe.get)(x_feat)
    # iv = ((df_woe_iv["p_bd"].to_numpy() - df_woe_iv["p_bd_"].to_numpy())*df_woe_iv["woe_pl"].to_numpy()).sum() # calculate information value
    return (woe_x_feat, descr_np)

def optimal_criter_split(arr_nan, left, right, criterion):
    """
    Define best split according to criterion (depends on nan allocation).
    Nan values add into turn branches (left and right), count all p-values.
    Choose minimal p-value and nan's allocation

    Parameters
    ----------
    arr_nan : structured array
        Contain censuring flag and time of events with feature missing values.
    left : structured array
        Contain censuring flag and time of left sample events.
    right : structured array
        Contain censuring flag and time of right sample events.
    criterion : function
        Weighted log-rank criteria.

    Returns
    -------
    min_p_val : float
        P-value of best split.
    none_to : int
        Number of branch for allocation (default 0)
        0 : to left branch
        1 : to right branch

    """
    none_to = 0
    min_p_val = 1.0
    if arr_nan.shape[1] > 0:
        left_and_nan = np.hstack([left, arr_nan])
        right_and_nan = np.hstack([right, arr_nan])
        a = criterion(left_and_nan[1], right[1], left_and_nan[0], right[0])
        b = criterion(left[1], right_and_nan[1], left[0], right_and_nan[0])
        # Nans move to a leaf with less p-value
        none_to = int(a > b)
        min_p_val = min(a,b)
    else:
        min_p_val = criterion(left[1], right[1], left[0], right[0])
    return (min_p_val, none_to)

def get_attrs(min_p_value, values, none_to, l_sh, r_sh, nan_sh):
    """
    Create dictionary of best split.

    Parameters
    ----------
    min_p_value : float
        P-value of best split.
    values : object
        Splitting value (can be single number or list).
    none_to : int
        Number of branch for allocation.
    l_sh : int
        Size of left branch.
    r_sh : int
        Size of right branch.
    nan_sh : int
        Quantitive of nan events.

    Returns
    -------
    attrs : dict
        Contain fields about best split.
        p_value : minimal p-value
        values : splitting value
        pos_nan : list of binary values of allocation nans
        min_split : minimal size of branch

    """
    attrs = {}
    attrs["p_value"] = min_p_value
    attrs["values"] = values
    if none_to:
        attrs["pos_nan"] = [0,1]
        attrs["min_split"] = min(l_sh,r_sh+nan_sh)
    else:
        attrs["pos_nan"] = [1,0]
        attrs["min_split"] = min(l_sh+nan_sh,r_sh)
    return attrs

def get_cont_attrs(uniq_set, arr_notnan, arr_nan, min_samples_leaf, criterion, 
                   signif, thres_cont_bin_max):
    """
    Find best split for continious feature.
    Consider all intermiate points of values (with quantile discretization).
    For each points define two branches and count p-value.
    Insignificant and too small splits aren't considered. 

    Parameters
    ----------
    uniq_set : set
        Unique values of feature.
    arr_notnan : structured array
        Contain censuring flag and time of events with feature missing values.
    arr_nan : structured array
        Contain feature value, censuring flag and time of events.
    min_samples_leaf : int
        Minimal acceptable size of branches.
    criterion : function
        Weighted log-rank criteria.
    signif : float
        Minimal acceptable significance of split.
    thres_cont_bin_max : int
        Maximal quantitive of intermiate points.

    Returns
    -------
    attr_dicts : list
        List of attr dicts (contain fields about best split).

    """
    if uniq_set.shape[0] > thres_cont_bin_max:
            uniq_set = np.quantile(arr_notnan[0], 
                    [i/float(thres_cont_bin_max) for i in range(1,thres_cont_bin_max)])
    else: # Set intermediate points
        uniq_set = (uniq_set[:-1] + uniq_set[1:])*0.5
    uniq_set = np.round(uniq_set, 3)
    attr_dicts = []
    for value in uniq_set:
        # Filter by attr value
        ind = arr_notnan[0]>=value
        left = arr_notnan[1:, np.where(ind)[0]].astype(np.int32)
        right = arr_notnan[1:, np.where(~ind)[0]].astype(np.int32)
        if min(left.shape[1],right.shape[1]) <= min_samples_leaf:
            continue
        min_p_value, none_to = optimal_criter_split(arr_nan, left, right, criterion)
        if min_p_value <= signif:
            attr_loc = get_attrs(min_p_value, value, none_to, 
                                 left.shape[1], right.shape[1], arr_nan.shape[1])
            attr_dicts.append(attr_loc)
    return attr_dicts

def get_categ_attrs(uniq_set, arr_notnan, arr_nan, min_samples_leaf, criterion, signif):
    """
    Find best split for categorical feature.
    Consider all nonoverlapping subsets of uniq elements.
    For each subset define branche and count p-value.
    Insignificant and too small splits aren't considered. 

    Parameters
    ----------
    uniq_set : set
        Unique values of feature.
    arr_notnan : structured array
        Contain censuring flag and time of events with feature missing values.
    arr_nan : structured array
        Contain feature value, censuring flag and time of events.
    min_samples_leaf : int
        Minimal acceptable size of branches.
    criterion : function
        Weighted log-rank criteria.
    signif : float
        Minimal acceptable significance of split.

    Returns
    -------
    attr_dicts : list
        List of attr dicts (contain fields about best split).

    """
    attr_dicts = []
    pairs_uniq = power_set_nonover(uniq_set)
    for l,r in pairs_uniq:
        left = arr_notnan[1:, np.isin(arr_notnan[0], l)].astype(np.int32)
        right = arr_notnan[1:, np.isin(arr_notnan[0], r)].astype(np.int32)
        if min(left.shape[1],right.shape[1]) <= min_samples_leaf:
            continue
        min_p_value, none_to = optimal_criter_split(arr_nan, left, right, criterion)
        if min_p_value <= signif:
            attr_loc = get_attrs(min_p_value, [list(l),list(r)], none_to, 
                                 left.shape[1], right.shape[1], arr_nan.shape[1])
            attr_dicts.append(attr_loc)
    return attr_dicts

def best_attr_split(arr, criterion = "logrank", type_attr = "cont", thres_cont_bin_max = 100,
              signif = 1.0, min_samples_leaf = 10, bonf = True, verbose = 0, 
              **kwargs):
    """
    Choose best split for fixed feature.
    Find best splits and choose partition with minimal p-value.

    Parameters
    ----------
    arr : structured array
        Contain feature value, censuring flag and time of events.
    criterion : object, optional
        Function or name of criteria. The default is "logrank".
        If str then replace to realization from criteria_dict
    type_attr : str, optional
        Continious, categorical or WOE features. The default is "cont".
        Continious : get_cont_attrs
        Categorical : get_categ_attrs
        WOE : mapping from categorical to continious and apply get_cont_attrs
    thres_cont_bin_max : int, optional
        Maximal quantitive of intermiate points.
    signif : TYPE, optional
        Minimal acceptable significance of split. The default is 1.0.
    min_samples_leaf : TYPE, optional
        Minimal acceptable size of branches. The default is 10.
    bonf : boolean, optional
        Apply Bonferroni adjustment. The default is True.
    verbose : int, optional
        Print best splits if verbose > 0. The default is 0.
    **kwargs : None
        Undefined parameters.

    Returns
    -------
    best_attr : dict
        Contain fields about best split.
        p_value : minimal p-value
        values : list of splitting rules
        pos_nan : list of binary values of allocation nans
        min_split : minimal size of branch
        sign_split : quantitive of significant splits

    """
    if criterion in scrit.CRITERIA_DICT:
        criterion = scrit.CRITERIA_DICT[criterion]
        
    best_attr = {"p_value":signif, "sign_split" : 0,
                  "values":[], "pos_nan":[1,0]}
    attr_dicts = [best_attr]
    # The leaf is too small for split
    if arr.shape[1] < 2*min_samples_leaf:
        return best_attr 
    ind = np.isnan(arr[0])
    arr_nan = arr[1:, np.where(ind)[0]].astype(np.int32)
    arr_notnan = arr[:, np.where(~ind)[0]]
    
    if type_attr == "woe":
        arr_notnan[0], descr_np = transform_woe(arr_notnan[0], arr_notnan[1])
    uniq_set = np.unique(arr_notnan[0])
    
    if type_attr == "categ" and uniq_set.shape[0] > 0:
        attr_dicts = get_categ_attrs(uniq_set, arr_notnan, arr_nan, min_samples_leaf, criterion, signif)
    else:
        attr_dicts = get_cont_attrs(uniq_set, arr_notnan, arr_nan, min_samples_leaf, criterion, signif, thres_cont_bin_max)
    
    if len(attr_dicts) == 0:
        return best_attr
    best_attr = min(attr_dicts,key=lambda x:x["p_value"])
    best_attr["sign_split"] = len(attr_dicts)
    if best_attr["sign_split"] > 0:
        if type_attr == "cont":
            best_attr["values"] = [' >= %s' % (best_attr["values"]), 
                                   ' < %s' % (best_attr["values"])]
        elif type_attr == "categ":
            best_attr["values"] = [' in %s' % (e) for e in best_attr["values"]]
        elif type_attr == "woe":
            ind = descr_np[1] >= best_attr["values"]
            l,r = list(descr_np[0, np.where(ind)[0]]), list(descr_np[0, np.where(~ind)[0]])
            best_attr["values"] = [' in %s' % (e) for e in [l,r]]
        # Bonferroni adjustment
        if bonf:
            best_attr["p_value"] *= best_attr["sign_split"]
        if verbose > 0:
            print(best_attr["p_value"], len(uniq_set))
    return best_attr


import os
import random
import numpy as np

RANDOM_STATE = 123
""" int: Fixed seed for model stability """

CENS_NAME = 'cens'
TIME_NAME = 'time'
""" str: Fixed names for internal representation """


def set_seed(seed_value):
    os.environ['PYTHONHASHSEED']=str(seed_value)
    random.seed(seed_value)
    np.random.seed(seed_value)


def get_y(cens, time):
    """
    Internal representation of target variable

    Parameters
    ----------
    cens : array-like, shape = (n_events,)
        Censoring flags.
    time : array-like, shape = (n_events,)
        Time of occured events.

    Returns
    -------
    y : structured array
        Output containing the binary event indicator as first field,
        and time of event or time of censoring as second field.

    """
    cens, time = np.array(cens), np.array(time) 
    y = np.empty(dtype=[(CENS_NAME, bool), 
                        (TIME_NAME, np.float64)], 
                 shape=cens.shape[0])
    y[CENS_NAME] = cens
    y[TIME_NAME] = time
    return y


def pd_to_xy(df):
    """
    Splitting pandas dataframe to feature space and target variables

    Parameters
    ----------
    df : pandas DataFrame
        Must to have columns CENS_NAME and TIME_NAME.

    Returns
    -------
    X : pandas DataFrame
        Feature space with remaining features.
    y : structured array
        y containing the binary event indicator as first field,
        and time of event or time of censoring as second field.

    """
    X = df.loc[:, set(df.columns) - {CENS_NAME, TIME_NAME}]
    y = get_y(df[CENS_NAME], df[TIME_NAME])
    return X, y


def get_bins(time, cens = None, mode = 'a', num_bins = 100):
    """
    Generate array of time points in timeline (from sample)

    Parameters
    ----------
    time : array-like
        Time of occured events.
    cens : array-like, optional
        Censoring flags. The default is None (all events occured).
    mode : str, optional
        Method of generation. The default is 'a'.
        'a' : all points (from min to max)
        'q' : quantile points (quantity is based on num_bins)
    num_bins : int, optional
        Quantity of required points. The default is 100.

    Returns
    -------
    bins : array
        Array of time points in timeline

    """
    if not(cens is None):
        time = time[np.where(cens)]
    time = time.astype(np.int32)
    bins = np.array([])
    if mode == 'q':
        bins = np.quantile(time, np.arange(num_bins) / num_bins)
    elif mode == 'a':
        bins = np.arange(time.min(), time.max()+1)
    return bins


import numpy as np
import pandas as pd
import re
import pickle 
import random
from os.path import dirname, join
from ..constants import TIME_NAME, CENS_NAME, get_y
random.seed(10)

descr_schemes = {
    '1':'Хирургическое_лечение',
    '2':'Лекарственная_противоопухолевая_терапия',
    '3':'Лучевая_терапия',
    '1_1':'Хирургическое_лечение_Первичной_опухоли_в_том_числе_с_удалением_регионарных_лимфатических_узлов',
    '1_2':'Хирургическое_лечение_Метастазов',
    '1_3':'Хирургическое_лечение_Симптоматическое_реконструктивно_пластическое_хирургическая_овариальная_суперссия_прочее',
    '1_4':'Хирургическое_лечение_Выполнено_хирургическое_стадирование',
    '1_5':'Хирургическое_лечение_Регионарных_лимфатических_узлов_без_первичной_опухоли',
    '1_6':'Хирургическое_лечение_Криохирургия_криотерапия_лазерная_деструкция_фотодинамическая_терапия',
    '2_1':'Лекарственная_противоопухолевая_терапия_Первая_линия',
    '2_2':'Лекарственная_противоопухолевая_терапия_Вторая_линия',
    '2_3':'Лекарственная_противоопухолевая_терапия_Третья_линия',
    '2_4':'Лекарственная_противоопухолевая_терапия_Линия_после_третьей',
    '2_5':'Лекарственная_противоопухолевая_терапия_Неоадъювантная',
    '2_6':'Лекарственная_противоопухолевая_терапия_Адъювантная',
    '2_7':'Лекарственная_противоопухолевая_терапия_Периоперационная_до_хирургического_лечения',
    '2_8':'Лекарственная_противоопухолевая_терапия_Периоперационная_после_хирургического_лечения',
    '3_1':'Лучевая_терапия_Первичной_опухоли_ложа_опухоли',
    '3_2':'Лучевая_терапия_Метастазов',
    '3_3':'Лучевая_терапия_Симптоматическая',
    '4':'Химиолучевая_терапия',
    '5':'Неспецифическое_лечение',
    '6':'Диагностика',
}

schemes_list = list(descr_schemes.keys())
schemes_bins =      ['BINS'+sch for sch in schemes_list]
schemes_nums =      ['NUM'+sch for sch in schemes_list]
schemes_pred_nums = ['PRED_NUM'+sch for sch in schemes_list]
schemes_pred_bins = ['PRED_BINS'+sch for sch in schemes_list]

sign = ['WEI', 'HEI', 'W', 'AGE',
        'STAD','ONK_T','ONK_N','ONK_M','MTSTZ', 'INDEX_MASS',
        'PRED_SUM', 'PRED_DAYS', 'PRED_MEAN_SUM_ELEMS',
        'PRED_MEAN_DAYS_ELEMS', 'PRED_MEAN_SUM_THREADS',
        'PRED_MEAN_DAYS_THREADS', 'DELTA_PRED', 'DELTA_POST', 'PRED_THREADS',
        'DIAG'
        ] + schemes_bins + schemes_nums + schemes_pred_bins + schemes_pred_nums

schemes_nan_fill = schemes_bins + schemes_nums
clustering_feature = ['PROFIL', 'SMO_NAM', 'SMO', 'ST_OKATO', 'CODE_MO','CODE_MD','PRVS','PODR','NPR_MO']

categ = ['DIAG']#,'DS2','DS3']#'STAD','VID_HMP','METOD_HMP',]
    

schemes_bins = {'BINS'+i:j for i,j in descr_schemes.items()}
schemes_nums = {'NUM'+i:'Кол_во_'+j for i,j in descr_schemes.items()}
schemes_pred_bins = {'PRED_BINS'+i:'Пред_кол_во_'+j for i,j in descr_schemes.items()}
schemes_pred_nums = {'PRED_NUM'+i:'Пред_лечение_'+j for i,j in descr_schemes.items()}

onk_descr = {
'WEI':'Вес',
'HEI':'Рост',
'W':'Пол',
'AGE':'Возраст',
'STAD':'Стадия',
'ONK_T':'Tumor_опухоль',
'ONK_N':'Nodus_узлы',
'ONK_M':"Метастазы",
'SUM_M':'Стоимость_лечения_в_рублях',
'INDEX_MASS':'Индекс_массы',
'PRED_SUM':'Предыдущая_стоимость_лечения',
'PRED_DAYS':'Длительность_прошлой_нити',
'PRED_MEAN_SUM_ELEMS':'Средняя_сумма_оплаченных_позиций',
'PRED_MEAN_DAYS_ELEMS':'Среднее_кол_во_дней',
'PRED_MEAN_SUM_THREADS':'Средняя_сумма_по_нитям',
'PRED_MEAN_DAYS_THREADS':'Средние_дни_по_нитям',
'DELTA_PRED':'Дни_между_прошлым_посещением',
'DELTA_POST':'Дни_между_прошлой_нитью',
'PRED_THREADS':'Предыдущая_нить',
'DIAG':'Диагноз',
'PROFIL':'Профиль',
'SMO_NAM':'Название_СК',
'SMO':'Номер_СК',
'ST_OKATO':'ОКАТО',
'CODE_MO':'Код_больницы',
'CODE_MD':'Код_врача',
'PRVS':'Специальность_работника',
'PODR':'Код_отделения',
'NPR_MO':'Код_МО_направившей_на_лечение'
}

onk_descr.update(schemes_bins)
onk_descr.update(schemes_nums)
onk_descr.update(schemes_pred_bins)
onk_descr.update(schemes_pred_nums)

STAD_MAP_DICT = {'0':0, 'I':1,'II':2,'III':3,'IV':4}

def get_int_str(s):
    s = re.sub(r'[^0-9]','',str(s))
    if len(s) == 0:
        return np.nan
    s = s[0]
    return int(s)

def prep_ONK(id_name, kod_name, namefile):
    ONK = pd.read_excel(namefile)
    ONK[kod_name] = ONK[kod_name].apply(lambda x: get_int_str(x))
    ONK = ONK.set_index(id_name)[kod_name].to_dict()
    return ONK

def save_pickle(obj, path):
    file_pi = open(path, 'wb') 
    pickle.dump(obj, file_pi, pickle.HIGHEST_PROTOCOL)
    
def load_pickle(path):
    return pickle.load(open(path, 'rb'))

def load_onk_dataset(diag = {'C20', 'C50.4', 'C61'}, 
                save_name = "", 
                invert_death = False,
                descript = True):
    dir_env = join(dirname(__file__), "data", "ONK")
    STAD_MAP = pd.read_excel(join(dir_env, 'STAD_MAP.xlsx'))
    STAD_MAP['KOD_St'] = STAD_MAP['KOD_St'].apply(lambda x: STAD_MAP_DICT.get(re.sub(r'[^IV0]','',x), np.nan))
    STAD_MAP = STAD_MAP.set_index('ID_St')['KOD_St'].to_dict()
    ### ONK PREP
    ONK_T = prep_ONK('ID_T', 'KOD_T', join(dir_env, 'ONK_T.xlsx'))
    ONK_N = prep_ONK('ID_N', 'KOD_N', join(dir_env, 'ONK_N.xlsx'))
    ONK_M = prep_ONK('ID_M', 'KOD_M', join(dir_env, 'ONK_M.xlsx'))
    
    sourceDF = pd.read_csv(join(dir_env, 'AGGREG_PRED_THREADS.csv'))
    if len(diag) > 0:
        sourceDF = sourceDF[sourceDF['DIAG'].isin(diag)]
        
    # if len(diag) > 0 and threadsDF[threadsDF['DIAG'] == diag].shape[0] > 0:
    #     if not os.path.exists(namespace.out +diag):
    #         os.mkdir(namespace.out +namespace.diag)
    #     namespace.out += namespace.diag + '\\'
    # #SAVE STAD
    # crl.save_pickle(STAD_MAP, namespace.out + 'STAD_MAPPING.pickle')
    # crl.save_pickle(ONK_T, namespace.out + 'ONK_T_MAPPING.pickle')
    # crl.save_pickle(ONK_N, namespace.out + 'ONK_N_MAPPING.pickle')
    # crl.save_pickle(ONK_M, namespace.out + 'ONK_M_MAPPING.pickle')
    
    sourceDF['STAD'] = sourceDF['STAD'].map(STAD_MAP)
    sourceDF['ONK_T'] = sourceDF['ONK_T'].map(ONK_T)
    sourceDF['ONK_N'] = sourceDF['ONK_N'].map(ONK_N)
    sourceDF['ONK_M'] = sourceDF['ONK_M'].map(ONK_M)
    if len(save_name) > 0:
        sourceDF.to_csv(join(dir_env, save_name), index = False)
    
    for i in categ:
        dict_encoder = {v_u:i_u for i_u, v_u in enumerate(sorted([y for y in sourceDF[i].unique() if y == y]))}
        # SAVE DICT
        #crl.save_pickle(dict_encoder, namespace.out + i + '.pickle')
        print(i, dict_encoder)
        sourceDF[i] = sourceDF[i].map(dict_encoder)
    
    if invert_death:
        sourceDF[CENS_NAME] = 1 - sourceDF['DEATH']
    else:
        sourceDF[CENS_NAME] = sourceDF['DEATH']
        
    sourceDF[TIME_NAME] = sourceDF['DAYS']
    
    ret_sign = sign
    ret_categ = categ
    ret_sch_nan = schemes_nan_fill
    if descript:
        repl = {i:onk_descr.get(i, i) for i in sign}
        sourceDF = sourceDF.rename(repl, axis = 1)
        
        ret_sign = [onk_descr.get(ns, ns) for ns in ret_sign]
        ret_categ = [onk_descr.get(ns, ns) for ns in ret_categ]
        ret_sch_nan = [onk_descr.get(ns, ns) for ns in ret_sch_nan]
        
    y = get_y(sourceDF[CENS_NAME], sourceDF[TIME_NAME])
    X = sourceDF.loc[:, ret_sign]
    return X, y, ret_sign, ret_categ, ret_sch_nan


import numpy as np
import pandas as pd

import matplotlib.pyplot as plt
import seaborn as sns
sns.set()

from sksurv.linear_model import CoxPHSurvivalAnalysis
from sksurv.tree import SurvivalTree
from sksurv.ensemble import RandomSurvivalForest
from sksurv.ensemble import GradientBoostingSurvivalAnalysis
    
from ..tree import CRAID
from ..ensemble import BootstrapCRAID
from ..ensemble import BoostingCRAID
# from .. import criteria as scrit 
from ..experiments import grid as exp
from .. import datasets as ds

from .PARAMS.GBSG_PARAM import GBSG_PARAMS
from .PARAMS.PBC_PARAM import PBC_PARAMS
from .PARAMS.ONK_PARAM import ONK_PARAMS
from .PARAMS.WUHAN_PARAM import WUHAN_PARAMS
from .PARAMS.COVID_PARAM import COVID_PARAMS

SELF_ALGS = {
    "TREE":CRAID,
    "BSTR":BootstrapCRAID,
    "BOOST":BoostingCRAID
}

PARAMS_ = {
    "GBSG": GBSG_PARAMS,
    "PBC": PBC_PARAMS,
    "ONK": ONK_PARAMS,
    "WUHAN": WUHAN_PARAMS,
    "COVID": COVID_PARAMS
}

DATASETS_LOAD = {
    "GBSG": ds.load_gbsg_dataset,
    "PBC": ds.load_pbc_dataset,
    "WUHAN": ds.load_wuhan_dataset,
    "ONK": ds.load_onk_dataset,
    "COVID": ds.load_covid_dataset
}

cox_param_grid = {'alpha':[100, 10, 1, 0.1, 0.01, 0.001], 
                  'ties':["breslow", "efron"]
                 }
RSF_param_grid = {'n_estimators':[30, 50, 100],
                  'max_depth':[None, 20],
                  'min_samples_leaf': [1, 10, 20], #[500, 1000, 3000],
                  'max_features':["sqrt"],
                  "random_state":[123]
                 }
ST_param_grid = {'splitter':["best", "random"], 
                       'max_depth':[None, 20, 30], 
                       'min_samples_leaf':[1, 10, 20],
                       'max_features':["sqrt"],
                       "random_state":[123]
                 }
GBSA_param_grid = {'loss':["coxph"], 
                  'learning_rate':[0.01, 0.05, 0.1, 0.5],
                  'n_estimators':[30, 50, 100], 
                  'max_depth':[20, 30], 
                  'min_samples_leaf':[1, 10, 20],
                  'max_features':["sqrt"],
                  "random_state":[123]
                 }

#[scrit.peto, scrit.wilcoxon, scrit.tarone_ware, scrit.logrank],

def get_best_by_full_name(df_full, by_metric = "IAUC", choose = "max"):
    df = df_full.copy()
    df['METHOD'] = df.apply(lambda x: x["METHOD"].replace("CRAID", "Tree(%s)" %(x['CRIT'])), axis = 1)
    if not(by_metric in df.columns):
        return None
    best_table = pd.DataFrame([], columns = df.columns)
    for method in df["METHOD"].unique():
        sub_table = df[df["METHOD"] == method]
        if sub_table.shape[0] == 0:
            continue
        if choose == "max":
            best_row = sub_table.loc[sub_table[by_metric].apply(np.mean).idxmax()]
        else:
            best_row = sub_table.loc[sub_table[by_metric].apply(np.mean).idxmin()]
        best_table = best_table.append(dict(best_row), ignore_index = True)
    return best_table

def plot_results(df_full, dir_path = None, metrics = [], 
                 dataset_name = "", all_best = False,
                 by_metric = "IAUC", choose = "max"):
    if not(all_best):
        df_ = get_best_by_full_name(df_full, by_metric, choose)
    for m in metrics:
        if all_best:
            df_ = get_best_by_full_name(df_full, m, choose = "min" if m == "IBS" else "max")
        plt.rcParams.update({'font.size': 15})
        fig, axs = plt.subplots(1, figsize=(8, 8))
        
        plt.title("%s %s" % (dataset_name, m))
        plt.boxplot(df_[m][::-1], labels = df_['METHOD'][::-1], showmeans = True, vert = False)
        if dir_path is None:
            plt.show()
        else:
            plt.savefig(dir_path + dataset_name + "%s_boxplot.png" %(m))
            plt.close(fig)
            
def import_tables(dirs):
    dfs = []
    for d in dirs:
        dfs.append(pd.read_excel(d))
    df = pd.concat(dfs)
    df = df.drop_duplicates()
    for c in ["IBS", "IAUC", "CI", "CI_CENS"]:
        df[c] = df[c].apply(lambda x: list(map(float,x[1:-1].split())))
    return df
    
def run(dataset = 'GBSG', with_self = ["TREE", "BSTR", "BOOST"], 
        with_external = True, except_stop = "all", dir_path = "./EXP_RES/"):
    """
    Conduct experiments for defined dataset and methods (self and external)

    Parameters
    ----------
    dataset : str, optional
        Name of dataset (must be in DATASETS_LOAD). The default is 'GBSG'.
    with_self : list, optional
        Names of self models. The default is ["TREE", "BSTR", "BOOST"].
        "TREE" : CRAID,
        "BSTR" : BootstrapCRAID,
        "BOOST" : BoostingCRAID.
    with_external : boolean, optional
        Flag of addiction external models. The default is True.
        Models : CoxPHSurvivalAnalysis, 
                 SurvivalTree,
                 RandomSurvivalForest,
                 GradientBoostingSurvivalAnalysis
    except_stop : str, optional
        Mode of ending because of exception. The default is "all".
        "all" - continue experiments
        else - stop experiments with current method
        
    dir_path : str, optional
        Path to the final directory

    Returns
    -------
    experim : Experiments
        Instance of class with constructed table.

    Examples
    --------
    from survivors.tests.experiment import run
    df_full, df_best = run()
    
    """
    lst_metrics = ["CI", "CI_CENS", "IBS", "IAUC"]
    if not(dataset in DATASETS_LOAD):
        print("DATASET %s IS NOT DEFINED" % (dataset))
    X, y, features, categ, sch_nan = DATASETS_LOAD[dataset]()
    experim = exp.Experiments(folds = 5, except_stop = except_stop, dataset_name = dataset)
    experim.set_metrics(lst_metrics)
    if with_external:
        experim.add_method(CoxPHSurvivalAnalysis, cox_param_grid)
        experim.add_method(SurvivalTree, ST_param_grid)
        experim.add_method(RandomSurvivalForest, RSF_param_grid)
        experim.add_method(GradientBoostingSurvivalAnalysis, GBSA_param_grid)
    if len(with_self) > 0:
        for alg in with_self:
            PARAMS_[dataset][alg]["categ"] = [categ]
            experim.add_method(SELF_ALGS[alg], PARAMS_[dataset][alg])
    experim.run(X, y, dir_path = dir_path, verbose = 1)
    return experim
    


short = False

CRAID_param_grid = {
    "depth": [15],
    "criterion": ["peto"] if short else ["peto", "tarone-ware", "wilcoxon", "logrank"],
    "min_samples_leaf": [5] if short else [5, 10, 20],
    'cut' : [True, False],
    "woe" : [False], #if short else [True, False], 
    "signif": [0.05] if short else [0.05, 0.15],
    "max_features": [1.0],
    "n_jobs" : [32]
}

BSTR_param_grid = {
    "size_sample": [0.7],
    "n_estimators": [10] if short else [10, 30],
    "depth": [15],
    "ens_metric_name": ["conc", "ibs"],
    # "woe" : [], 
    "criterion": ["peto"] if short else ["peto", "tarone-ware", "wilcoxon", "logrank"], 
    "min_samples_leaf": [10, 20] if short else [5, 10],
    "max_features": [0.5] if short else [0.5, "sqrt"],
    "n_jobs" : [32]
}

# BOOST_param_grid = {
#     "size_sample": [0.5] if short else [0.7],
#     "n_estimators": [15] if short else [15], 
#     "ens_metric_name": ["ibs"] if short else ["conc","ibs"],
#     "depth": [15, 25],
#     "mode_wei": ['exp'] if short else ['square','exp'],
#     # "woe" : [],
#     "criterion": ["logrank"] if short else ["peto", "tarone-ware", "wilcoxon", "logrank"],
#     "min_samples_leaf": [5, 10, 20],
#     "max_features": [0.7] if short else [0.7, "sqrt"],
#     "aggreg_func": ['wei'] if short else ['wei', 'mean']
# }

BOOST_param_grid = {
    "size_sample": [0.5] if short else [0.7],
    "n_estimators": [15] if short else [20], 
    "ens_metric_name": ["ibs"] if short else ["conc","ibs"],
    "depth": [15],
    "mode_wei": ['exp'] if short else ['exp', 'linear'],
    # "woe" : [],
    "criterion": ["logrank"] if short else ["peto", "tarone-ware", "wilcoxon", "logrank"],
    "min_samples_leaf": [10, 25],
    "max_features": [0.7] if short else [0.7],
    "aggreg_func": ['wei'] if short else ['wei', 'mean'],
    "n_jobs" : [32]
}


WUHAN_PARAMS = {
    "TREE": CRAID_param_grid,
    "BSTR": BSTR_param_grid,
    "BOOST": BOOST_param_grid
}